#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

import base64
import json
from typing import Dict, Text, Any

import apache_beam as beam
import numpy as np
import tensorflow as tf
from dateutil.parser import parse

from zenml.core.components.data_gen import constants
from zenml.core.components.data_gen.constants import BINARY_DATA, FILE_NAME, \
    BASE_PATH, FILE_PATH, FILE_EXT, METADATA, IMAGE, LABEL
from zenml.core.components.data_gen.constants import DestinationKeys
from zenml.core.components.data_gen.types.interface import TypeInterface
from zenml.core.components.data_gen.utils import init_bq_table, DtypeInferrer
from zenml.core.components.data_gen.utils import is_smaller_than


@beam.ptransform_fn
@beam.typehints.with_input_types(beam.typehints.Dict[Text, Any])
@beam.typehints.with_output_types(beam.pvalue.PDone)
def WriteImagesToBQ(datapoints: Dict[Text, Any],
                    schema: Dict[Text, Any],
                    destination: Dict[Text, Any]) -> beam.pvalue.PDone:
    """
    Args:
        datapoints:
        schema:
        destination:
    """
    allowed_ext = [".jpg", ".json", ".png", ".txt", ".jpeg"]

    images, label_file = (datapoints
                          | "FilterOutFiles" >> beam.Filter(lambda x:
                                                            x[FILE_EXT] in
                                                            allowed_ext)
                          | "SplitLabelFile" >> beam.Partition(SplitByFileName,
                                                               2))

    # label_file is actually a dict
    label_dict = beam.pvalue.AsSingleton(label_file)

    ready_images = (images
                    | "AddLabelAndMetadata" >> beam.Map(add_label_and_metadata,
                                                        label_dict)
                    | "AddTFExample" >> beam.Map(append_tf_example_to_image)
                    )

    # Obtain the schema
    if schema:
        schema_dict = schema
        assert all(v in ['STRING', 'INTEGER', 'FLOAT', 'BYTES']
                   for _, v in schema.items())
        init_bq_table(
            destination[DestinationKeys.PROJECT],
            destination[DestinationKeys.DATASET],
            destination[DestinationKeys.TABLE],
            schema)
    else:
        schema = (ready_images
                  | 'InferSchema' >> beam.CombineGlobally(
                    DtypeInferrer(),
                    project=destination[DestinationKeys.PROJECT],
                    dataset=destination[DestinationKeys.DATASET],
                    table=destination[DestinationKeys.TABLE]))

        schema_dict = beam.pvalue.AsSingleton(schema)

    # filter out data points larger than 10 MiB (default)
    final_images = (ready_images
                    | "FilterBySize" >> beam.Filter(is_smaller_than)
                    )

    result = (final_images
              | "WriteToBQ" >> beam.io.WriteToBigQuery(
                table=destination[DestinationKeys.TABLE],
                dataset=destination[DestinationKeys.DATASET],
                project=destination[DestinationKeys.PROJECT],
                write_disposition='WRITE_EMPTY',
                create_disposition="CREATE_NEVER")
              )

    return result


def _int64_feature(value):
    """Returns an int64_list from a bool / enum / int / uint.

    Args:
        value:
    """
    input_value = value.flatten() if (hasattr(value, 'shape') and
                                      (len(value.shape) > 1)) else [value]
    return tf.train.Feature(int64_list=tf.train.Int64List(value=input_value))


def _float_feature(value):
    """Returns an int64_list from a bool / enum / int / uint.

    Args:
        value:
    """
    input_value = value.flatten() if (hasattr(value, 'shape') and
                                      (len(value.shape) > 1)) else [value]
    return tf.train.Feature(float_list=tf.train.FloatList(
        value=input_value))


def _bytes_feature(value):
    """Returns a bytes_list from a string / byte.

    Args:
        value:
    """
    if isinstance(value, type(tf.constant(0))):
        value = value.numpy()
    return tf.train.Feature(bytes_list=tf.train.BytesList(value=[value]))


def _label_feature(value):
    """Returns an int64_list from a bool / enum / int / uint.

    Args:
        value:
    """
    # flatten first if it is a label array, which is the case e.g. in
    # computer vision + segmentation
    input_value = value.flatten() if (hasattr(value, 'shape') and
                                      (len(value.shape) > 1)) else [value]

    return tf.train.Feature(int64_list=tf.train.Int64List(
        value=input_value))


def _metadata_feature(metadata):
    # Returns all the metadata in the image as a string
    """
    Args:
        metadata:
    """
    metadata_as_bytes = str.encode(metadata)
    return tf.train.Feature(bytes_list=tf.train.BytesList(
        value=[metadata_as_bytes]))


def get_matching_label(label_data: Text, img_filename: Text):
    """
    Args:
        label_data (Text): Should be JSON-readable.
        img_filename (Text):

    Returns:
        * **label** (*Str, label key*)
        * **metadata** (*Dict, additional metadata information*)
    """

    # TODO: This can potentially go bad if the JSON is too big
    data_dict = json.loads(label_data)

    img_data = data_dict[img_filename]

    if "label" not in img_data:
        raise ValueError("Error while parsing label file: No label "
                         "key was found. Make sure to specify your "
                         "image label under a \"label\" key.")

    label = img_data.pop(LABEL)

    try:
        metadata = img_data.pop(METADATA)
    except KeyError:
        metadata = {}

    # JSON parsing puts empty metadata dict as the empty string,
    # so in that case coerce to dict just to be type safe
    if not metadata:
        metadata = {}

    return label, metadata


def add_label_and_metadata(image_dict: Dict[Text, Any], label_dict: Text):
    """Parse a ByteString Image and convert to TFExample

    Args:
        image_dict:
        label_dict (Text):
    """
    filename = image_dict[FILE_NAME]

    label_data = label_dict[BINARY_DATA]
    label, metadata = get_matching_label(label_data, filename)

    image_dict[METADATA] = metadata
    image_dict[LABEL] = label

    return image_dict


def parse_metadata(metadata_dict: Dict[Text, Any]):
    """
    Args:
        metadata_dict:
    """
    key_feature_map = {
        "int": _int64_feature,
        "float": _float_feature,
        "str": _metadata_feature,
        "bytes": _bytes_feature}

    md_features = {}
    new_metadata = metadata_dict.copy()

    # separate check if value is list-like, also numpy.ndarray coercion
    for key, value in metadata_dict.items():
        val_type = type(value)
        if isinstance(value, list):
            value = np.array(value).flatten()
            val_type = type(value[0])

        # check whether a str is actually a timestamp,
        # important so that BQ does not throw an error
        if isinstance(value, str):
            try:
                ts = parse(value, fuzzy=True)
                ts = ts.strftime('%Y-%m-%dT%H:%M:%S.%f %Z')
                new_metadata[key] = ts
            except:
                pass

        md_features[key] = key_feature_map.get(val_type.__name__)(value)

    return new_metadata, md_features


def append_tf_example_to_image(image_dict: Dict[Text, Any]):
    # pop the binary data, base_path, metadata
    """
    Args:
        image_dict:
    """
    image_raw = image_dict.pop(BINARY_DATA)
    base_path = image_dict.pop(BASE_PATH)
    file_path = image_dict.pop(FILE_PATH)
    metadata_dict = image_dict.pop(METADATA)

    parsed_metadata, md_feature = parse_metadata(metadata_dict)
    # image = tf.io.decode_image(image_raw).numpy()
    # image_shape = image.shape

    feature = {
        # 'height': _int64_feature(image_shape[0]),
        # 'width': _int64_feature(image_shape[1]),
        # 'depth': _int64_feature(image_shape[2]),
        IMAGE: _bytes_feature(image_raw),
        LABEL: _label_feature(image_dict[LABEL]),
    }

    # update feature dict with metadata information
    feature.update(md_feature)
    # update image_dict with metadata info
    image_dict.update(parsed_metadata)

    tf_example = tf.train.Example(features=tf.train.Features(feature=feature))

    image_dict[constants.DATA_COL] = base64.b64encode(
        tf_example.SerializeToString())

    return image_dict


def SplitByFileName(element: Dict[Text, Any],
                    num_partitions: int) -> int:
    """
    Args:
        element:
        num_partitions (int):
    """
    try:
        # filename = element[FILE_NAME]
        file_ext = element[FILE_EXT]

        if file_ext.lower() in [".txt", ".json", ".csv"]:
            return 1
    # if we get a KeyError here, we are likely processing a BQ dict,
    # which should have the label already inside
    except KeyError:
        return 0

    return 0


class ImageConverter(TypeInterface):
    def convert(self):
        return WriteImagesToBQ



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Entrypoint for gcp orchestrator"""

import base64
import json

import absl
import fire

from zenml.core.backends.orchestrator.local.orchestrator_local_backend import \
    OrchestratorLocalBackend


class PipelineRunner(object):

    def run_pipeline(self, config_b64: str):
        # Load config from base64
        config = json.loads(base64.b64decode(config_b64))

        # Change orchestrator of pipeline to local
        OrchestratorLocalBackend().run(config)


if __name__ == "__main__":
    # execute only if run as a script
    absl.logging.set_verbosity(absl.logging.INFO)
    fire.Fire(PipelineRunner)



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

import json
import os
from math import isclose
from typing import List, Text, Dict, Any

import absl
from google.cloud import bigquery

from zenml.core.components.data_gen.constants import DATA_COL
from zenml.core.components.split_gen import constants
from zenml.core.pipelines.config.standard_config import SplitKeys, GlobalKeys
from zenml.utils.constants import ZENML_PIPELINE_CONFIG


def check_split_ratio_sum(split_ratio_dict: Dict[Text, Any]):
    # using round because of float inaccuracies
    if round(sum(split_ratio_dict.values()), 15) != 1:
        raise AssertionError("Split ratios must sum up to 1.")


def check_split_ratio_keys(split_ratio_dict: Dict[Text, Any]):
    """
    Small helper to validate a split ratio dictionary by its keys.
    Allowed: Either train + eval, train + eval + test, or nosplit keys.
    """
    # Get fold names
    split_folds = list(split_ratio_dict.keys())
    num_folds = len(split_folds)

    required_folds = []

    if not split_folds:
        raise AssertionError("Error: The given split ratio "
                             "dictionary is empty.")

    # nosplit case
    if num_folds == 1:
        required_folds = [constants.NOSPLIT]

    elif num_folds == 2:
        required_folds = [constants.TRAIN, constants.EVAL]

    elif num_folds == 3:
        required_folds = [constants.TRAIN, constants.EVAL, constants.TEST]

    elif num_folds > 3:
        raise AssertionError(
            "Error: More than three fold keys are not supported at the "
            "moment.")

    if not required_folds:
        raise AssertionError(
            "Error: Something went wrong while validating the split ratio "
            "dict."
            "This should not happen. Contact a maiot GmbH engineer "
            "immediately!")

    if not all(str(fold) in required_folds for fold in split_folds):
        raise AssertionError(
            "Error while validating the given ratio dict. Only the following "
            "fold names are supported "
            "when specifying {0} folds: ".format(num_folds)
            + '[%s]' % ', '.join(required_folds))


def validate_split_ratio_dict(split_ratio_dict: Dict[Text, Any]):
    check_split_ratio_sum(split_ratio_dict)
    check_split_ratio_keys(split_ratio_dict)


def validate_split_config(split_config):
    """
    Helper to validate the supplied split config in terms of its entries.
    This is to prohibit some pathological cases (can be expanded).
    """
    # ratio dict checks for vertical and horizontal split,
    # top-level ratio checks happen below
    for key in [SplitKeys.CATEGORIZE_BY_, SplitKeys.INDEX_BY_]:
        if key in split_config:
            split = split_config[key]
            if constants.RATIO in split:
                ratio_dict = split[constants.RATIO]
                validate_split_ratio_dict(ratio_dict)

    # top-level ratio dict exists
    if SplitKeys.RATIO_ in split_config:

        # keys in the top-level ratio dict
        top_level_ratio_dict = split_config[SplitKeys.RATIO_]

        # split ratio dict checks (sum + key consistency)
        validate_split_ratio_dict(top_level_ratio_dict)
        top_level_folds = list(top_level_ratio_dict.keys())

        # disallow all three ratio dicts to be set simultaneously
        if SplitKeys.CATEGORIZE_BY_ in split_config:
            # get the config for horizontal split
            categorize_by_cfg = split_config[SplitKeys.CATEGORIZE_BY_]
            if constants.RATIO in categorize_by_cfg:
                category_ratio_dict = categorize_by_cfg[constants.RATIO]

                categorize_by_folds = list(
                    category_ratio_dict.keys())
                # set comprehension to validate ratio dicts
                if not set(categorize_by_folds) == set(top_level_folds):
                    # Print the keys creating the error for verbosity
                    # to help the user
                    set_diff = (set(categorize_by_folds) - set(
                        top_level_folds)) \
                               or (set(top_level_folds) - set(
                        categorize_by_folds))
                    raise AssertionError(
                        "Error: Ratio dict inconsistency detected. Please "
                        "check that you "
                        "supply the same keys for each split ratio entry in "
                        "the config. Keys that do not appear in both ratio"
                        "dicts: {}".format(set_diff))

            if constants.CATEGORIES in categorize_by_cfg:
                categories = categorize_by_cfg[constants.CATEGORIES]
                # disallow category specification and top-level split at the
                # same time
                if isinstance(categories, dict):
                    raise AssertionError(
                        "Error: Specifying a top-level split ratio and a "
                        "category dict at the same time is "
                        "disallowed.")

            if SplitKeys.INDEX_BY_ in split_config:
                # get the config for vertical split
                index_by_cfg = split_config[SplitKeys.INDEX_BY_]

                # disallow all ratio dicts to be set simultaneously
                if all(constants.RATIO in cfg for cfg in
                       [index_by_cfg, categorize_by_cfg]):
                    raise AssertionError(
                        "Error: Illegal ratio specification detected."
                        "Either drop the top-level \'ratio\' config key or "
                        "the lower-level index-by \'ratio\' key.")

        # disallow top-level ratio dict and index_by ratio dict to be set
        # simultaneously
        if SplitKeys.INDEX_BY_ in split_config:
            index_by_cfg = split_config[SplitKeys.INDEX_BY_]
            if constants.RATIO in index_by_cfg:
                raise AssertionError("Illegal ratio specification detected. "
                                     "Specifying a top-level split ratio "
                                     "and an index-by ratio at the "
                                     "same time is not supported.")

    else:  # no top-level ratio dict
        if SplitKeys.CATEGORIZE_BY_ in split_config:
            # get the config for horizontal split
            categorize_by_cfg = split_config[SplitKeys.CATEGORIZE_BY_]
            if constants.CATEGORIES in categorize_by_cfg:
                categories = categorize_by_cfg[constants.CATEGORIES]
                if isinstance(categories, dict):
                    if constants.RATIO in categorize_by_cfg:
                        # disallow category specification and category split at
                        # the same time
                        raise AssertionError(
                            "Error: Specifying a category split ratio and a "
                            "category dict "
                            "at the same time is disallowed.")

                    if constants.NOSPLIT in categories:
                        # handle the case where nosplit is specified along with
                        # other options as an error
                        if len(categories.keys()) > 1:
                            raise AssertionError(
                                "Error: When specifying \'nosplit\', passing "
                                "other fold keys is disallowed.")

                elif isinstance(categories, list):
                    if constants.RATIO not in categorize_by_cfg:
                        raise AssertionError(
                            "Error: When passing a list, the category split "
                            "ratio has to be specified,"
                            "either by a top-level or by a "
                            "category-level ratio dict.")

            else:
                # no top-level or category ratio keys specified
                if constants.RATIO not in categorize_by_cfg:
                    raise AssertionError(
                        "Error: When passing a list, the category split ratio "
                        "has to be specified,"
                        "either by a top-level or "
                        "by a category-level ratio dict.")

            if SplitKeys.INDEX_BY_ in split_config:
                # get the config for vertical split
                index_by_cfg = split_config[SplitKeys.INDEX_BY_]
                if all(constants.RATIO in cfg for cfg in
                       [index_by_cfg, categorize_by_cfg]):

                    category_ratio_dict = categorize_by_cfg[constants.RATIO]
                    idx_ratio_dict = index_by_cfg[constants.RATIO]

                    categorize_by_folds = list(
                        category_ratio_dict.keys())
                    index_by_folds = list(idx_ratio_dict.keys())
                    # set comprehension to validate ratio dicts
                    if not set(categorize_by_folds) == set(index_by_folds):
                        # Print the keys creating the error for verbosity
                        # to help the user
                        set_diff = (set(categorize_by_folds) - set(
                            index_by_folds)) or \
                                   (set(index_by_folds) - set(
                                       categorize_by_folds))
                        raise AssertionError(
                            "Error: Ratio dict inconsistency detected. Please "
                            "check that you supply the same keys for "
                            "each split ratio entry in the config. "
                            "Keys that do not appear in both ratio "
                            "dicts: {}".format(set_diff))

                if constants.RATIO in index_by_cfg:
                    idx_ratio_dict = index_by_cfg[constants.RATIO]

                    index_by_folds = list(idx_ratio_dict.keys())
                    if constants.CATEGORIES in categorize_by_cfg:
                        categories = categorize_by_cfg[constants.CATEGORIES]
                        if isinstance(categories, dict):
                            # set comprehension to validate category dict
                            # and index ratio dict
                            categorize_by_folds = list(categories.keys())
                            if not set(categorize_by_folds) == set(
                                    index_by_folds):
                                # Print the keys creating the error
                                # for verbosity to help the user
                                set_diff = (set(categorize_by_folds) - set(
                                    index_by_folds)) \
                                           or (set(index_by_folds) - set(
                                    categorize_by_folds))
                                raise AssertionError(
                                    "Error: Ratio dict inconsistency "
                                    "detected. Please check that you "
                                    "supply the same keys for each split"
                                    " ratio entry in the config. Keys that"
                                    " do not appear in both ratio "
                                    "dicts: {}".format(set_diff))
                        elif isinstance(categories, list):
                            if constants.RATIO not in categorize_by_cfg:
                                raise AssertionError(
                                    "Error: When passing a list, the category"
                                    " split ratio has to be specified,"
                                    "either by a top-level or by a "
                                    "category-level ratio dict.")


def get_cat_len_map(exec_properties):
    """
    Initial query to get count of data per category
    :return: dict of type { category: num_datapoints_in_category }
    """
    client = bigquery.Client()

    bq_args = extract_bq_args(exec_properties)
    cat_col = exec_properties[constants.CAT_COL]

    # TODO: [HIGH] Since the categories (eval and train) are hashed, we
    #  have to derive them from the config every time, remove this very
    #  ugly hack
    category_dict = infer_cats(bq_args=bq_args, split_config=None)

    # Dummy query to get the type information for each field.
    if cat_col:
        query = 'SELECT {cat_col}, COUNT({cat_col}) as count FROM ' \
                '`{project}.{dataset}.{table}` '.format(
            project=bq_args['project'],
            table=bq_args['table'],
            dataset=bq_args['dataset'],
            cat_col=cat_col)
        query = handle_where(query,
                             category_column=exec_properties[
                                 constants.CAT_COL],
                             category_dict=category_dict,
                             extras=exec_properties[constants.WHERE])
        query = query + ' GROUP BY {cat_col}'.format(cat_col=cat_col)
        query_job = client.query(query)
        return {row[cat_col]: row['count'] for row in query_job.result()}
    else:
        query = 'SELECT COUNT(*) as count FROM ' \
                '`{project}.{dataset}.{table}`'.format(
            project=bq_args['project'],
            dataset=bq_args['dataset'],
            table=bq_args['table'])
        query = handle_where(query,
                             category_column=exec_properties[
                                 constants.CAT_COL],
                             category_dict=category_dict,
                             extras=exec_properties[constants.WHERE])
        query_job = client.query(query)
        # Just need to loop for dtype consistency
        return {constants.DEFAULT_CATEGORY: row['count']
                for row in query_job.result()}


# TODO: This needs a complete rewrite, bisect does not do what it is
#  designed for in this setting
def get_limit_map(exec_properties):
    id_len_map = get_cat_len_map(exec_properties)

    # TODO: Hardcoded variables :(
    possible_idx_ratio_keys = [constants.I_RATIO_TRAIN,
                               constants.I_RATIO_EVAL,
                               constants.I_RATIO_TEST,
                               constants.I_RATIO_NOSPLIT]

    conversion_dict = {constants.I_RATIO_TRAIN: constants.TRAIN,
                       constants.I_RATIO_EVAL: constants.EVAL,
                       constants.I_RATIO_TEST: constants.TEST,
                       constants.I_RATIO_NOSPLIT: constants.NOSPLIT}

    limit_map = {}
    for k, v in id_len_map.items():
        ratio = 0
        limit_dict = {}
        for i_ratio_key in possible_idx_ratio_keys:
            i_ratio = exec_properties[i_ratio_key]
            if i_ratio is not None:
                ratio += i_ratio
                converted_key = conversion_dict[i_ratio_key]
                # this check is an artifact induced by the bisect.bisect
                # function in our custom Beam PartitionFn.
                # The right endpoint of the last fold (equal to the total
                # number of data points) in the limit dict
                # needs to be shifted by one so that bisect does not
                # create another split fold, which it would do for the
                # last row because the buckets are
                # half-open intervals B = [a,b).
                if isclose(ratio, 1.0):
                    limit_dict[converted_key] = v * ratio + 1
                else:
                    limit_dict[converted_key] = v * ratio

        limit_map[k] = limit_dict

    return limit_map


def extract_bq_args(exec_prop):
    return {'project': exec_prop[constants.BQ_PROJECT],
            'dataset': exec_prop[constants.BQ_DATASET],
            'table': exec_prop[constants.BQ_TABLE]}


def get_possible_categories(bq_args, cat_col):
    client = bigquery.Client()
    query_job = client.query(
        'SELECT DISTINCT ({cat_col}) as category '
        'FROM `{project}.{dataset}.{table}`'.format(
            project=bq_args['project'],
            table=bq_args['table'],
            dataset=bq_args['dataset'],
            cat_col=cat_col))
    return [row['category'] for row in query_job.result()]


def handle_features(features):
    if features is None:
        query_features = '*'
    elif isinstance(features, list):
        # In case we extend the standard for "defaults" reflect that here:
        query_features = features.copy()
        query_features = ",".join(list(set(query_features)))
    else:
        raise Exception(
            'You reached what I thought to be an unreachable state. '
            'I could give you advice what to do, but why would you trust me? '
            'Clearly I screwed this up - I am writing a message that should '
            'never appear, and yet it just did. '
            'I should stop coding and become a carpenter. Sorry!')
    return query_features


def handle_where(query,
                 category_column,
                 category_dict,
                 extras):
    where_clauses = []

    if extras is not None:
        harmful_sql = ['drop', 'update', 'truncate', 'commit', 'select',
                       'insert', 'create', 'alter']
        for clause in extras:
            if any(x in clause.lower().split() for x in harmful_sql):
                raise AssertionError('Dangerous SQL in WHERE '
                                     'clause: {}'.format(clause))
            where_clauses.append(clause)

    if any(len(cat_list) > 0 for cat_list in category_dict.values()):
        categories = []
        for cat_list in category_dict.values():
            categories = categories + cat_list  # add train, eval,
            # test category lists
        categories_str = [str(x) for x in categories]
        if type(categories[0]) is str:
            # if its string we wrap it around double quotes
            categories_joined = '"' + '","'.join(categories_str) + '"'
        else:
            # if numeric we dont wrap
            categories_joined = ','.join(categories_str)

        clause = '{column} in ({categories})'.format(
            column=category_column,
            categories=categories_joined)
        where_clauses.append(clause)

    if len(where_clauses):
        query += 'WHERE '
        query += " and ".join(where_clauses)

    return query


def parse_query(exec_properties, ce_config=None):
    bq_args = extract_bq_args(exec_properties)

    cat_col = exec_properties[constants.CAT_COL]
    idx_col = exec_properties[constants.IDX_COL]

    cols_to_select = [DATA_COL]

    cat_query = ''
    if cat_col:
        cols_to_select.append(cat_col)
        cat_query = 'PARTITION BY {}'.format(cat_col)

    idx_query = ''
    if idx_col:
        cols_to_select.append(idx_col)
        idx_query = 'ORDER BY {}'.format(idx_col)

    query = '''SELECT {features}, 
               ROW_NUMBER() OVER ({cat_query} {idx_query}) as {dfn}
               FROM `{project}.{dataset}.{table}`
               '''.format(
        features=','.join(list(set(cols_to_select))),
        project=bq_args['project'],
        dataset=bq_args['dataset'],
        table=bq_args['table'],
        cat_query=cat_query,
        idx_query=idx_query,
        dfn=constants.DEFAULT_ROW_NUMBER)

    category_dict = infer_cats(bq_args, ce_config)

    query = handle_where(query,
                         category_column=cat_col,
                         category_dict=category_dict,
                         extras=exec_properties[constants.WHERE])

    if exec_properties[constants.TEST_MODE]:
        query = query + ' LIMIT {}'.format(constants.LIMIT)

    absl.logging.warning(query)
    return query


def infer_cats(bq_args, split_config):
    if split_config is None:
        split_config = json.loads(os.getenv(ZENML_PIPELINE_CONFIG))[
            GlobalKeys.SPLIT]

    cat_dict = {}

    if SplitKeys.CATEGORIZE_BY_ in split_config:
        categorize_by_cfg = split_config[SplitKeys.CATEGORIZE_BY_]

        if constants.CATEGORIES in categorize_by_cfg:
            categories = categorize_by_cfg[constants.CATEGORIES]
            if isinstance(categories, dict):
                # sort the category lists and then that's it
                cat_dict = {k: sorted(l) for k, l in categories.items()}

            elif isinstance(categories, list):
                categories = sorted(categories)

                if SplitKeys.RATIO_ in split_config:
                    c_ratio = split_config[SplitKeys.RATIO_]
                else:
                    c_ratio = categorize_by_cfg[constants.RATIO]

                cat_dict = partition_cat_list(categories, c_ratio)

    # map fold keys to category dict keys
    cat_converter = {constants.TRAIN: constants.CAT_TRAIN,
                     constants.EVAL: constants.CAT_EVAL,
                     constants.TEST: constants.CAT_TEST,
                     constants.NOSPLIT: constants.CAT_NOSPLIT}

    mapped_dict = {cat_converter[k]: v for k, v in cat_dict.items()}

    return mapped_dict


def map_cat_to_fold(cat, cat_map) -> int:
    """
    Small helper to map folds to integers in the Partition Function.
    """
    folds = [constants.CAT_TRAIN, constants.CAT_EVAL, constants.CAT_TEST,
             constants.CAT_NOSPLIT]

    for i, fold in enumerate(folds):
        if fold in cat_map:
            fold_list = cat_map[fold]
            if not isinstance(fold_list, list):
                raise AssertionError("Error: Expected list but got element "
                                     "of type {}".format(type(fold_list)))
            if cat in fold_list:
                # this hack puts nosplit as fold zero, avoiding
                # problems with the Beam partition function
                # (in our workloads, nosplit does not appear together
                # with other folds, so this should be fine)
                return i % 3

    # case where the category is not in any fold
    return -1


def partition_cat_list(cat_list: List[Text], c_ratio: Dict[Text, Any]):
    """
    Helper to split a category list by the entries in a category split dict.
    Scales well in the number of folds
    """
    cat_dict = {}

    num_cats = len(cat_list)
    ratio = 0

    # This might produce unexpected results if the number of categories
    # is lower than the number of folds.
    for fold, fold_pct in c_ratio.items():
        left_bound = round(num_cats * ratio)
        ratio += fold_pct
        right_bound = round(num_cats * ratio)
        # write categories for fold to the cat_list dict
        cat_dict[fold] = cat_list[left_bound:right_bound]

    return cat_dict


def infer_split_names(exec_properties):
    """
    Small function to infer the split names from a dict passed to the TFX
    component executor.
    """
    # TODO: Hardcoded folds :( later: folds = exec_properties[
    #  constants.SPLIT_NAMES] or something
    cat_options = [constants.CAT_TRAIN, constants.CAT_EVAL, constants.CAT_TEST,
                   constants.CAT_NOSPLIT]
    c_ratio_options = [constants.C_RATIO_TRAIN, constants.C_RATIO_EVAL,
                       constants.C_RATIO_TEST, constants.C_RATIO_NOSPLIT]
    idx_options = [constants.I_RATIO_TRAIN, constants.I_RATIO_EVAL,
                   constants.I_RATIO_TEST, constants.I_RATIO_NOSPLIT]
    names = [constants.TRAIN, constants.EVAL, constants.TEST,
             constants.NOSPLIT]

    # rather obscure dict construction to map cat and idx constants to the
    # split names (workloads)
    map_dict = dict(zip(cat_options + idx_options + c_ratio_options,
                        names + names + names))

    split_names = []
    if not any(exec_properties[opt] is not None
               for opt in cat_options + c_ratio_options):
        # separate check! categorical and index can exist independently
        if not any(exec_properties[opt] is not None for opt in idx_options):
            # "naked config", corresponds to nosplit
            split_names.append(constants.NOSPLIT)
            return split_names
        else:
            # pick up all the split folds from the idx portion
            for opt in idx_options:
                if exec_properties[opt] is not None:
                    split_names.append(map_dict[opt])
            return split_names

    else:
        for opt in cat_options:
            if exec_properties[opt] is not None:
                split_names.append(map_dict[opt])

        for opt in c_ratio_options:
            if exec_properties[opt] is not None:
                split_names.append(map_dict[opt])
        # Here we assume that categorical and index ratio dicts are consistent,
        # i.e. have the same keys and workload (train-eval, nosplit etc.)
        return split_names


def get_category_map(exec_properties):
    """
    Helper to infer the categories at executor runtime, either from the
    plaintext entries in the CAT_xyz entries or from the C_RATIO_xyz entries
    in the exec_properties.
    """

    mapped_dict = {}

    c_ratio_keys = [constants.C_RATIO_TRAIN, constants.C_RATIO_EVAL,
                    constants.C_RATIO_TEST, constants.C_RATIO_NOSPLIT]

    cat_keys = [constants.CAT_TRAIN, constants.CAT_EVAL,
                constants.CAT_TEST, constants.CAT_NOSPLIT]

    if any(exec_properties[key] is not None for key in c_ratio_keys):
        # extract BQ args and categorical column
        bq_args = extract_bq_args(exec_properties)

        cat_col = exec_properties[constants.CAT_COL]

        # infer categories by dummy BQ query
        categories = sorted(get_possible_categories(bq_args, cat_col))

        c_ratio = {key: exec_properties[key] for key in c_ratio_keys
                   if exec_properties[key] is not None}

        # partition the category list with the ratio in the executor spec
        cat_dict = partition_cat_list(categories, c_ratio)

        # map fold keys to category dict keys
        cat_converter = dict(zip(c_ratio_keys, cat_keys))

        mapped_dict = {cat_converter[k]: v for k, v in cat_dict.items()}

    if any(exec_properties[key] is not None for key in cat_keys):
        # exec_properties are concatenated lists of categories with comma as
        # separator -> split them by the "," inside
        bq_args = extract_bq_args(exec_properties)

        cat_col = exec_properties[constants.CAT_COL]

        # Query to infer cat types
        categories = sorted(get_possible_categories(bq_args, cat_col))

        # TODO: Can it happen that multiple categories out of a BigQuery job
        #  have different dtypes? Then this is not good
        cat_type = type(categories[0])

        cat_dict = {key: exec_properties[key].split(",") for key in cat_keys
                    if exec_properties[key] is not None}

        for key, cat_list in cat_dict.items():
            cat_dict[key] = [cat_type(cat) for cat in cat_list]

        mapped_dict = cat_dict

    return mapped_dict



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

import json
import os
from textwrap import dedent
from zenml.utils.path_utils import read_file_contents

import click


# IMPORT FUNCTIONS
def parse_metrics(d):
    tmp = d.copy()
    m = tmp.pop('metrics')
    for k, v in m.items():
        tmp['metric_' + k] = v['doubleValue']
    return tmp


def get_log_dir(p_uuid, r_uuid, info):
    # TODO: how do i kow that the pipeline is in this workspace, maybe i
    #   changed the ws
    ws_id = info[info[constants.ACTIVE_USER]][constants.ACTIVE_WORKSPACE]
    d_path = os.path.join(click.get_app_dir(constants.APP_NAME),
                          'eval_trainer',
                          str(ws_id),
                          str(p_uuid),
                          str(r_uuid))

    if os.path.exists(os.path.join(d_path, 'eval_model_dir')):
        return d_path

    api = ce_api.PipelinesApi(api_client(info))
    artifact = api_call(
        api.get_pipeline_artifacts_api_v1_pipelines_pipeline_id_runs_pipeline_run_id_artifacts_component_type_get,
        pipeline_id=p_uuid,
        pipeline_run_id=r_uuid,
        component_type=GDPComponent.Trainer.name)
    download_artifact(artifact[0].to_dict(), path=d_path)
    return d_path


def get_eval_dir(p_uuid, r_uuid, info, d_path=None):
    ws_id = info[info[constants.ACTIVE_USER]][constants.ACTIVE_WORKSPACE]

    if d_path is None:
        d_path = os.path.join(click.get_app_dir(constants.APP_NAME),
                              'eval_evaluator',
                              str(ws_id),
                              str(p_uuid),
                              str(r_uuid))

    if os.path.exists(os.path.join(d_path, 'eval_config.json')):
        return d_path

    api = ce_api.PipelinesApi(api_client(info))
    artifact = api_call(
        api.get_pipeline_artifacts_api_v1_pipelines_pipeline_id_runs_pipeline_run_id_artifacts_component_type_get,
        pipeline_id=p_uuid,
        pipeline_run_id=r_uuid,
        component_type=GDPComponent.Evaluator.name)

    # TODO: [LOW] artifact[1] hard-coded because of upgrade to 0.21.4
    download_artifact(artifact[0].to_dict(), path=d_path)

    # replace google path with local path
    with open(os.path.join(d_path, 'eval_config.json'), 'r') as f:
        eval_config = json.load(f)

    # now override the google path to local path
    eval_config['modelLocations'][''] = d_path

    with open(os.path.join(d_path, 'eval_config.json'), 'w') as f:
        json.dump(eval_config, f)

    return d_path


# THE REAL STUFF
def import_block():
    block = '''\
        import param
        import panel as pn
        import pandas as pd
        import plotly.express as px
        import tensorflow_model_analysis as tfma
        import ce_api
        from ce_cli import constants
        from ce_cli.utils import api_call, api_client, download_artifact, 
        notice, \
            declare, format_uuid, find_closest_uuid
        import plotly.graph_objects as go

        from ce_cli.evaluation import parse_metrics, get_eval_dir

        pn.extension('plotly')
    '''
    return dedent(block)[:-1]


def info_block(info):
    block = '''\
        info = {}
    '''.format(info)
    return dedent(block)[:-1]


def application_block():
    block = '''\
        class Application(param.Parameterized):
            pipeline_run_selector = param.ListSelector(default=[], objects=[
            ]) # 1
            hyperparameter_selector = param.ListSelector(default=[], 
            objects=[]) # 2

            slicing_metric_selector = param.ObjectSelector(default='', 
            objects=['']) # 3
            performance_metric_selector = param.ObjectSelector(objects=[]) # 4 


            def __init__(self, **params):
                super(Application, self).__init__(**params)
                # apis
                p_api = ce_api.PipelinesApi(api_client(info))
                ws_api = ce_api.WorkspacesApi(api_client(info))

                user = info[constants.ACTIVE_USER]
                active_workspace = info[user][constants.ACTIVE_WORKSPACE]

                # lists 
                result_list = []
                hparam_list = []

                # get all pipelines in this workspace
                all_pipelines = api_call(
                ws_api.get_workspaces_pipelines_api_v1_workspaces_workspace_id_pipelines_get, 
                                         active_workspace)

                # get a dataframe of all results + all hyperparameter 
                combinations
                for p in all_pipelines:
                    if p.pipeline_runs is not None and len(p.pipeline_runs)>0:
                        for run in p.pipeline_runs:
                            # This is slowing the comparison down but 
                            necessary to update the status of each run
                            r = api_call(
                                p_api.get_pipeline_run_api_v1_pipelines_pipeline_id_runs_pipeline_run_id_get,
                                p.id,
                                run.id)
                            if r.status == 'Succeeded':
                                eval_path = get_eval_dir(p.id, r.id, info)
                                evaluation = tfma.load_eval_result(eval_path)
                                for s, m in evaluation.slicing_metrics:
                                    result_list.append(dict([(
                                    'pipeline_name', '{}:{}'.format(p.name, 
                                    r.id)),
                                                             ('slice_name', 
                                                             s[0][0] if s 
                                                             else ''), 
                                                             ('slice_value', 
                                                             s[0][1] if s 
                                                             else ''), 
                                                             ('metrics', 
                                                             m[''][''])]))

                                h_dict = api_call(
                                    p_api.get_hyperparameters_pipeline_api_v1_pipelines_pipeline_id_runs_pipeline_run_id_hyperparameters_get, 
                                    p.id,
                                    r.id)
                                h_dict['pipeline_name'] = '{}:{}'.format(
                                p.name, r.id)
                                hparam_list.append(h_dict)

                self.results = pd.DataFrame([parse_metrics(r) for r in 
                result_list])
                self.hparam_info = pd.DataFrame(hparam_list)

                # set params
                self.param.pipeline_run_selector.objects = self.results[
                'pipeline_name'].unique()


            def generate_results_with_additional_params(self, df, 
            track_list, pipeline_run_selector):
                print('Generating results...')
                tracking_dict = dict()
                for key in track_list:
                    component, parameter = key.split(':')
                    if component in tracking_dict:
                        tracking_dict[component].append(parameter)
                    else:
                        tracking_dict[component] = [parameter]

                context_list = list(self.results[self.results[
                'pipeline_name'].isin(pipeline_run_selector)][
                'context_id'].unique())
                for c in context_list:
                    for e in api_get_list_of_executions(int(c), info)[0]:
                        component_id = e['properties']['component_id'][
                        'stringValue']
                        if component_id in tracking_dict:
                            parameter_list = tracking_dict[component_id]
                            for p in parameter_list:
                                param_name = ':'.join([component_id, p])
                                param_value = e['properties'][p]['stringValue']
                                df.loc[df['context_id'] == c, param_name] = 
                                float(param_value)

                return df


            @param.depends('pipeline_run_selector', watch=True)
            def _updated_context(self):
                df = self.results[self.results['pipeline_name'].isin(
                self.pipeline_run_selector)]
                df = df.dropna(axis=1, how='all')

                slicing_metric_list = sorted(list(df['slice_name'].unique()))

                performance_metric_set = {c for c in df.columns if 
                c.startswith('metric_')}
                performance_metric_list = [None] + sorted(list(
                performance_metric_set))

                self.param['slicing_metric_selector'].objects = 
                slicing_metric_list
                self.param['performance_metric_selector'].objects = 
                performance_metric_list

                # get params
                parameter_list = self.hparam_info[self.hparam_info[
                'pipeline_name'].isin(self.pipeline_run_selector)].columns
                parameter_list = [x for x in parameter_list if x != 
                'pipeline_name']
                self.param['hyperparameter_selector'].objects = parameter_list

                self.slicing_metric_selector = ''
                self.performance_metric_selector = None


            @param.depends('slicing_metric_selector', 
            'performance_metric_selector', watch=True)
            def performance_graph(self): 
                if self.performance_metric_selector:
                    df = self.results[(self.results['pipeline_name'].isin(
                    self.pipeline_run_selector)) & 
                                      (self.results['slice_name'] == 
                                      self.slicing_metric_selector)]
                    fig = px.scatter(df,
                                     x='pipeline_name',
                                     y=self.performance_metric_selector,
                                     color='slice_value',
                                     width=1100,
                                     title='Pipeline Comparison')

                    fig = fig.update_traces(mode='lines+markers')

                else:
                    fig = px.scatter(pd.DataFrame(),
                                     marginal_y='rug',
                                     width=1100,
                                     title='Pipeline Comparison')

                return fig


            @param.depends('performance_metric_selector', 
            'hyperparameter_selector', watch=True)
            def parameter_graph(self): 
                if self.performance_metric_selector and len(
                self.hyperparameter_selector) > 0:
                    df = self.results[(self.results['pipeline_name'].isin(
                    self.pipeline_run_selector)) & 
                                      (self.results['slice_name'] == '')]

                    # merge
                    extra_df = pd.merge(self.hparam_info, df, 
                    on='pipeline_name', how='left')

                    dimensions = ['pipeline_name'] + 
                    self.hyperparameter_selector + [
                    self.performance_metric_selector]

                    new_dims = []
                    for d in dimensions:
                        try:
                            new_dims.append({'label': d, 'values': 
                            pd.to_numeric(extra_df[d])})
                        except:
                            u = sorted(list(extra_df[d].apply(lambda x: str(
                            x)).unique()))
                            mapping = {v: i for i, v in enumerate(u)}
                            new_dims.append({'label': d, 
                                             'tickvals': [mapping[x] for x 
                                             in u],
                                             'ticktext': u,
                                             'values': extra_df[d].apply(
                                             lambda x: str(x)).map(mapping)})

                    final_col = pd.to_numeric(extra_df[
                    self.performance_metric_selector])
                    fig = go.Figure(data=go.Parcoords(line=dict(
                    color=final_col,
                                                                colorscale = 
                                                                'inferno',
                                                                showscale=True,
                                                                cmin=min(
                                                                final_col),
                                                                cmax=max(
                                                                final_col)),
                                                      dimensions=new_dims))
                else:
                    fig = px.scatter(pd.DataFrame(),
                                     marginal_y='rug',
                                     width=1100,
                                     title='Hyperparameter Comparison')

                return fig
    '''
    return dedent(block)[:-1]


def interface_block():
    block = '''\
        def generate_interface():

            app = Application()
            handlers = pn.Param(app.param)

            # Analysis Page
            analysis_page = pn.GridSpec(height = 850, width=1850, max_height 
            = 850, max_width=1850)
            analysis_page[0:8, 0:2] = handlers[1]
            analysis_page[0:10, 8:10] = handlers[2]
            analysis_page[8:9, 0:2] = handlers[3]
            analysis_page[9:10, 0:2] = handlers[4]
            analysis_page[0:5, 2:8] = app.performance_graph
            analysis_page[5:10, 2:8] = app.parameter_graph

            interface = pn.Tabs(
                ('Analysis Page', analysis_page),
            )
            return interface

        platform = generate_interface()
        platform.servable()
    '''
    return dedent(block)[:-1]


def get_compare_block():
    return dedent(read_file_contents('compare.py'))[:-1]


def get_tensorboard_block(log_dir):
    block = '''\
    import os

    model_path = '"{evaluation}"'
    logdir = os.path.join(model_path, 'logs')
    %load_ext tensorboard
    '''.format(evaluation=log_dir)
    block = dedent(block)
    block += '%tensorboard --logdir {logdir}'
    return block


def get_eval_block(eval_dir):
    block = '''\
    # If the visualization does not appear after running this block, please 
    # run the same cell again

    import tensorflow_model_analysis as tfma

    evaluation_path = '{evaluation}'
    evaluation = tfma.load_eval_result(output_path=evaluation_path)

    # find slicing metrics
    slicing_columns = set([x.feature_keys[0] for x in 
    evaluation.config.slicing_specs if len(x.feature_keys) == 1])
    print("")
    print("Available slicing columns: ")
    print(slicing_columns)

    tfma.view.render_slicing_metrics(evaluation)

    # in order to view sliced results, pass in the `slicing_column` parameter:
    # tfma.view.render_slicing_metrics(evaluation, slicing_column='col_name')
    '''.format(evaluation=eval_dir)
    return dedent(block)[:-1]



DEFAULT_CAT = '_DEFAULT_'

#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

# GLOBAL execution parameters
RESAMPLING_RATE = 'resampling_rate_in_secs'
TRIP_GAP_THRESHOLD = 'trip_gap_threshold_in_secs'
TIMESTAMP_COLUMN = 'timestamp_column'
CATEGORY_COLUMN = 'category_column'

# PRETRANSFORM specific
FEATURES = 'features'
EVAL_FEATURES = 'eval_features'
LABELS = 'labels'
RESAMPLING = 'resampling'
FILLING = 'filling'
LABEL_TUNING = 'label_tuning'

DEFAULT_PREFIX = '_default'

# SEQUENCEGEN specific
SEQUENCE_LENGTH = 'sequence_length'
SEQUENCE_SHIFT = 'sequence_shift'



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

from zenml.core.pipelines.config.config_utils import ConfigKeys


class GlobalKeys(ConfigKeys):
    VERSION = 'version'
    ENV = 'environment'
    STEPS = 'steps'
    DATASOURCE = 'datasource'


class EnvironmentKeys(ConfigKeys):
    EXPERIMENT_NAME = 'pipeline_name'
    ARTIFACT_STORE = 'artifact_store'
    METADATA_STORE = 'metadata'
    ENABLE_CACHE = 'enable_cache'
    BACKENDS = 'backends'


class DatasourceKeys(ConfigKeys):
    NAME = 'name'
    SOURCE = 'source'
    ID = 'id'


class BackendKeys(ConfigKeys):
    TYPE = 'type'
    ARGS = 'args'


class TrainingSteps(ConfigKeys):
    DATA = 'data'
    SPLIT = 'split'
    SEQUENCING = 'sequencing'
    PREPROCESSING = 'preprocessing'
    TRAINING = 'training'
    EVALUATION = 'evaluation'
    DEPLOYMENT = 'deployment'


class DataSteps(ConfigKeys):
    DATA = 'data'


class InferSteps(ConfigKeys):
    DATA = 'data'
    INFER = 'infer'


# HELPER KEYS

class MethodKeys(ConfigKeys):
    METHOD = 'method'
    PARAMETERS = 'parameters'



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

from typing import Dict, Text, Any, List, Union

from zenml import Split, Index, Categorize


def no_split() -> Split:
    """Helper function to create a no-split configuration"""
    return Split()


def index_split(by: Text,
                ratio: Dict[Text, Any] = None) -> Split:
    """Helper function to create a indexed split configuration

    Args:
        by (Text): name of the column to index by
        ratio: dict to define the splits, e.g. {"train":0.8, "eval":0.2}

    Returns:
        an instance of Split with the proper configuration
    """
    return Split(index=Index(by=by, ratio=ratio))


def category_split(by: Text,
                   ratio: Dict[Text, Any] = None,
                   categories: Union[Dict, List] = None) -> Split:
    """Helper function to create a categorical split configuration

    Args:
        by (Text): name of the column to categorize by
        ratio: dict to define the splits, e.g. {"train":0.8, "eval":0.2}
        categories: dict or list to define the categories

    Returns:
        an instance of Split with the proper configuration
    """
    return Split(categorize=Categorize(by=by,
                                       ratio=ratio,
                                       categories=categories))


def hybrid_split(index_by: Text,
                 category_by: Text,
                 index_ratio: Dict[Text, Any] = None,
                 category_ratio: Dict[Text, Any] = None) -> Split:
    """Helper function to create a hybrid split configuration with both index
    and categorical elements

    Args:
        index_by (Text): name of the column to index by
        category_by (Text): name of the column to categorize by
        index_ratio: dict to define the index split ratios
        category_ratio: dict to define the category split ratios

    Returns:
        an instance of Split with the proper configuration
    """
    return Split(index=Index(by=index_by, ratio=index_ratio),
                 categorize=Categorize(by=category_by, ratio=category_ratio))



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

import math
import os
from math import isclose
from typing import List, Text, Dict, Any

import tensorflow as tf
import tensorflow_data_validation as tfdv
from tensorflow_data_validation.types import FeaturePath
from tensorflow_data_validation.utils import stats_util
from tfx import types
from tfx.components.schema_gen.executor import _DEFAULT_FILE_NAME, SCHEMA_KEY
from tfx.types import artifact_utils
from tfx.types.artifact import Artifact
from tfx.utils import io_utils

from zenml.core.components.split_gen import constants


def infer_split_names(exec_properties):
    """
    Small function to infer the split names from a dict passed to the TFX
    component executor.
    """
    # TODO: Hardcoded folds :( later: folds = exec_properties[
    #  constants.SPLIT_NAMES] or something
    cat_options = [constants.CAT_TRAIN, constants.CAT_EVAL, constants.CAT_TEST,
                   constants.CAT_NOSPLIT]
    c_ratio_options = [constants.C_RATIO_TRAIN, constants.C_RATIO_EVAL,
                       constants.C_RATIO_TEST, constants.C_RATIO_NOSPLIT]
    idx_options = [constants.I_RATIO_TRAIN, constants.I_RATIO_EVAL,
                   constants.I_RATIO_TEST, constants.I_RATIO_NOSPLIT]
    names = [constants.TRAIN, constants.EVAL, constants.TEST,
             constants.NOSPLIT]

    # rather obscure dict construction to map cat and idx constants to the
    # split names (workloads)
    map_dict = dict(zip(cat_options + idx_options + c_ratio_options,
                        names + names + names))

    split_names = []
    if not any(exec_properties[opt] is not None
               for opt in cat_options + c_ratio_options):
        # separate check! categorical and index can exist independently
        if not any(exec_properties[opt] is not None for opt in idx_options):
            # "naked config", corresponds to nosplit
            split_names.append(constants.NOSPLIT)
            return split_names
        else:
            # pick up all the split folds from the idx portion
            for opt in idx_options:
                if exec_properties[opt] is not None:
                    split_names.append(map_dict[opt])
            return split_names

    else:
        for opt in cat_options:
            if exec_properties[opt] is not None:
                split_names.append(map_dict[opt])

        for opt in c_ratio_options:
            if exec_properties[opt] is not None:
                split_names.append(map_dict[opt])
        # Here we assume that categorical and index ratio dicts are consistent,
        # i.e. have the same keys and workload (train-eval, nosplit etc.)
        return split_names


def construct_limit_map(cat_len_map: Dict[Text, int],
                        category_map: Dict[Text, List[Text]],
                        index_split: Dict[Text, float]):
    folds = [constants.TRAIN, constants.EVAL,
             constants.TEST, constants.NOSPLIT]
    limit_map = {}

    for i, cat_list in enumerate(category_map.values()):
        for cat in cat_list:
            limit_map[cat] = dict(zip(folds[:i], [-math.inf] * i))

    for k, v in cat_len_map.items():
        if k in limit_map:
            continue

        ratio = 0
        limit_dict = {}
        for fold, index_ratio in index_split.items():
            ratio += index_ratio

            if not isclose(ratio, 1.0):
                limit_dict[fold] = v * ratio

        limit_map[k] = limit_dict

    return limit_map


def parse_statistics(split_name: Text,
                     statistics: List[Artifact]) -> Dict[Text, int]:
    stats_uri = io_utils.get_only_uri_in_dir(
        artifact_utils.get_split_uri(statistics, split_name))

    stats = tfdv.load_statistics(stats_uri)

    return stats


def parse_schema(input_dict: Dict[Text, List[types.Artifact]]):
    schema = input_dict.get(SCHEMA_KEY, None)
    if not schema:
        return schema
    else:
        schema_path = os.path.join(
            artifact_utils.get_single_uri(schema),
            _DEFAULT_FILE_NAME)
        schema_reader = io_utils.SchemaReader()
        parsed_schema = schema_reader.read(schema_path)

        return parsed_schema


def parse_cat_len_map(statistics,
                      category_column: Text):
    feature_stats = stats_util.get_feature_stats(statistics.datasets[0],
                                                 FeaturePath(
                                                     [category_column]))
    if not hasattr(feature_stats, 'string_stats'):
        # this is a problem
        raise AssertionError(
            "You can only do horizontal splits on categorical variables")

    category_map = {}
    for v in feature_stats.string_stats.top_values:
        key = v.value
        n = v.frequency
        category_map[key] = n

    return category_map


def get_index_split(split_names, exec_properties) -> Dict[Text, float]:
    index_split = dict()
    # converter for fold keys to index ratio fold keys
    idx_ratio_converter = {constants.TRAIN: constants.I_RATIO_TRAIN,
                           constants.EVAL: constants.I_RATIO_EVAL,
                           constants.TEST: constants.I_RATIO_TEST,
                           constants.NOSPLIT: constants.I_RATIO_NOSPLIT}

    for name in split_names:
        converted_name = idx_ratio_converter[name]
        if exec_properties[converted_name]:
            index_split[name] = exec_properties[converted_name]

    return index_split


def get_category_map(schema, exec_properties):
    """
    Helper to infer the categories at executor runtime, either from the
    plaintext entries in the CAT_xyz entries or from the C_RATIO_xyz entries
    in the exec_properties.
    """

    mapped_dict = {}

    c_ratio_keys = [constants.C_RATIO_TRAIN, constants.C_RATIO_EVAL,
                    constants.C_RATIO_TEST, constants.C_RATIO_NOSPLIT]

    cat_keys = [constants.CAT_TRAIN, constants.CAT_EVAL,
                constants.CAT_TEST, constants.CAT_NOSPLIT]

    if any(exec_properties[key] is not None for key in c_ratio_keys):
        # extract categorical column
        cat_col = exec_properties[constants.CAT_COL]

        # infer categories by dummy BQ query
        categories = sorted(get_possible_categories(schema, cat_col))

        c_ratio = {key: exec_properties[key] for key in c_ratio_keys
                   if exec_properties[key] is not None}

        # partition the category list with the ratio in the executor spec
        cat_dict = partition_cat_list(categories, c_ratio)

        # map fold keys to category dict keys
        cat_converter = dict(zip(c_ratio_keys, cat_keys))

        mapped_dict = {cat_converter[k]: v for k, v in cat_dict.items()}

    if any(exec_properties[key] is not None for key in cat_keys):
        # exec_properties are concatenated lists of categories with comma as
        # separator -> split them by the "," inside
        cat_col = exec_properties[constants.CAT_COL]

        cat_dict = {key: exec_properties[key].split(",") for key in cat_keys
                    if exec_properties[key] is not None}

        mapped_dict = cat_dict

    return mapped_dict


def partition_cat_list(cat_list: List[Text], c_ratio: Dict[Text, Any]):
    """
    Helper to split a category list by the entries in a category split dict.
    Scales well in the number of folds
    """
    cat_dict = {}

    num_cats = len(cat_list)
    ratio = 0

    # This might produce unexpected results if the number of categories
    # is lower than the number of folds.
    for fold, fold_pct in c_ratio.items():
        left_bound = round(num_cats * ratio)
        ratio += fold_pct
        right_bound = round(num_cats * ratio)
        # write categories for fold to the cat_list dict
        cat_dict[fold] = cat_list[left_bound:right_bound]

    return cat_dict


def get_possible_categories(schema, cat_col: Text):
    category_domain = tfdv.get_domain(schema, cat_col)

    # get category list from result proto
    value_list = list(category_domain.value)

    return value_list


def get_categorical_value(example: tf.train.Example, cat_col: Text):
    cat_feature = example.features.feature[cat_col]

    possible_types = ["bytes", "float", "int64"]

    for datatype in possible_types:
        value_list = getattr(cat_feature, datatype + "_list")

        if value_list.value:
            value = value_list.value[0]
            if hasattr(value, "decode"):
                return value.decode()
            else:
                return value

    # TODO: Should we raise here? This could happen in a NULL type situation
    return None



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

import os
from typing import Dict, Text, Any

from tfx.extensions.google_cloud_ai_platform.trainer.executor import \
    TRAINING_ARGS_KEY
from tfx.utils import io_utils
from zenml.utils.standards import standard_dataline as std_dl
from zenml.utils.standards import standard_environment as std_env
from zenml.utils.standards import standard_inferline as std_if

from zenml.cli.version import PACKAGE_VERSION, PACKAGE_NAME
from zenml.utils.constants import TRAINER_IMAGE
from zenml.utils.constants import TrainingTypes


def get_pipeline_spec(env_dict):
    # Checking the global keys
    """
    Args:
        env_dict:
    """
    check_config(env_dict)

    # Dataset, experiment
    experiment_name = env_dict[std_env.GlobalKeys.EXPERIMENT_NAME]
    run_name = generate_unique_name(experiment_name)

    # Output
    output_base_dir = env_dict[std_env.GlobalKeys.OUTPUT_BASE_DIR]
    if not output_base_dir.startswith('gs://'):
        output_base_dir = os.path.abspath(output_base_dir)

    # Pipeline settings
    pipeline_name = experiment_name
    pipeline_enable_cache = env_dict[std_env.GlobalKeys.ENABLE_CACHE]
    project_id = env_dict[std_env.GlobalKeys.PROJECT_ID]
    gcp_region = env_dict[std_env.GlobalKeys.GCP_REGION]

    # Pipeline directories
    pipeline_root = os.path.join(output_base_dir, experiment_name)
    pipeline_log = os.path.join(pipeline_root, 'logs', run_name)
    pipeline_temp = os.path.join(pipeline_root, 'tmp', run_name)
    serving_model_dir = os.path.join(pipeline_root, 'serving_model', run_name)

    flags_to_add = {'project': project_id,
                    'region': gcp_region,
                    'temp_location': pipeline_temp,
                    'staging_location': pipeline_temp}

    # Orchestrator
    orch = env_dict[std_env.GlobalKeys.ORCHESTRATION][
        std_env.OrchestratorKeys.TYPE]
    orch_args = env_dict[std_env.GlobalKeys.ORCHESTRATION][
        std_env.OrchestratorKeys.ARGS]

    api_args = {}
    if orch == 'beam':
        orch_args['setup_file'] = os.path.join(os.getcwd(), 'setup.py')
        orch_args['job_name'] = 'gdp-' + run_name
        orch_args.update(flags_to_add)
        orch_args = parse_yaml_beam_args(orch_args)
    elif orch == 'gcp':
        pass
    elif orch == 'ce':
        api_args = env_dict[std_env.GlobalKeys.API_ARGS_]
        orch_args['setup_file'] = os.path.join(os.getcwd(), 'setup.py')
        orch_args['job_name'] = 'gdp-' + run_name
        orch_args.update(flags_to_add)
        orch_args = parse_yaml_beam_args(orch_args)
    elif orch == 'kubeflow':
        # TODO: Kubeflow specific stuff here
        pass
    else:
        raise AssertionError('Unknown orchestrator!')

    # BEPA
    exec_args = env_dict[std_env.GlobalKeys.EXECUTION]
    dist_path = os.path.join(os.getcwd(), 'dist')

    # Need to add empty requirements because of TFX bug in dependency_utils.py
    req_path = os.path.join(dist_path, 'requirements.txt')
    io_utils.write_string_file(req_path, '')
    exec_args['requirements_file'] = req_path

    gz_path = os.path.join(dist_path, '{}-{}.tar.gz'.format(
        PACKAGE_NAME,
        PACKAGE_VERSION))
    exec_args['extra_package'] = gz_path

    exec_args.update(flags_to_add)
    exec_args = parse_yaml_beam_args(exec_args)

    # BigQuery Settings
    try:
        bq_args = env_dict[std_env.GlobalKeys.BQ_ARGS]
        if 'project' not in bq_args:
            bq_args['project'] = project_id
        if 'table' not in bq_args:
            bq_args['table'] = experiment_name
    except AttributeError:
        bq_args = {}

    # Metadata Store
    metadata_connection_config = create_metadata_connection_config(
        env_dict[std_env.GlobalKeys.METADATA_ARGS],
        pipeline_name,
        output_base_dir)

    # Training
    training = env_dict[std_env.GlobalKeys.TRAINING]

    training_type = training[std_env.TrainingKeys.TYPE]
    training_args = training[std_env.TrainingKeys.ARGS]

    ai_platform_training_args = {
        'project': project_id,
        'region': gcp_region,
        'jobDir': pipeline_temp,
        'masterConfig': {
            'imageUri':
                TRAINER_IMAGE
        }
    }

    if training_type == TrainingTypes.gcaip.name:
        ai_platform_training_args['project'] = project_id

        ai_platform_training_args['scaleTier'] = training_args[
            std_env.AIPlatformKeys.SCALE_TIER]

        if std_env.AIPlatformKeys.RUNTIME_VERSION_ in training_args:
            ai_platform_training_args['runtimeVersion'] = training_args[
                std_env.AIPlatformKeys.RUNTIME_VERSION_]

        if std_env.AIPlatformKeys.PYTHON_VERSION_ in training_args:
            ai_platform_training_args['pythonVersion'] = training_args[
                std_env.AIPlatformKeys.PYTHON_VERSION_]

        if std_env.AIPlatformKeys.MAX_RUNNING_TIME_ in training_args:
            ai_platform_training_args['scheduling'] = {
                'maxRunningTime': training_args[
                    std_env.AIPlatformKeys.MAX_RUNNING_TIME_]}

    # Serving
    serving = env_dict[std_env.GlobalKeys.SERVING]

    serving_type = serving[std_env.ServingKeys.TYPE]
    ai_platform_serving_args = {
        'model_name': 'model_' + experiment_name.replace('-', '_'),
        'project_id': project_id
    }

    test_mode = env_dict[std_env.GlobalKeys.TEST_MODE]

    return {
        'pipeline_name': pipeline_name,
        'pipeline_root': pipeline_root,
        'pipeline_enable_cache': pipeline_enable_cache,
        'pipeline_log_root': pipeline_log,
        TRAINING_ARGS_KEY: ai_platform_training_args,
        'ai_platform_serving_args': ai_platform_serving_args,
        'metadata_connection_config': metadata_connection_config,
        'orchestrator': orch,
        'orchestration_args': orch_args,
        'execution_args': exec_args,
        'bq_args': bq_args,
        'training_type': training_type,
        'serving_type': serving_type,
        'serving_model_dir': serving_model_dir,
        'test_mode': test_mode,
        std_env.GlobalKeys.API_ARGS_: api_args,
    }


def get_dataline_spec(dataline_config: Dict[Text, Any]):
    """
    Args:
        dataline_config:
    """
    datasource = dataline_config[std_dl.DatalineKeys.DATASOURCE]
    destination = dataline_config[std_dl.DatalineKeys.DESTINATION]
    schema = dataline_config[std_dl.DatalineKeys.SCHEMA_]

    # Pipeline settings
    pipeline_name = dataline_config[std_dl.DatalineKeys.EXPERIMENT_NAME]
    pipeline_enable_cache = dataline_config[std_dl.DatalineKeys.ENABLE_CACHE]
    project_id = dataline_config[std_dl.DatalineKeys.GCP_PROJECT]
    gcp_region = dataline_config[std_dl.DatalineKeys.GCP_REGION]

    # Pipeline directories
    output_base_dir = dataline_config[std_dl.DatalineKeys.OUTPUT_BASE_DIR]
    if not output_base_dir.startswith('gs://'):
        output_base_dir = os.path.abspath(output_base_dir)

    pipeline_root = os.path.join(output_base_dir, pipeline_name)
    run_name = generate_unique_name(pipeline_name)
    pipeline_log = os.path.join(pipeline_root, 'logs', run_name)
    pipeline_temp = os.path.join(pipeline_root, 'tmp', run_name)

    # Orchestrator
    orch = dataline_config[std_dl.DatalineKeys.ORCHESTRATION][
        std_dl.OrchestratorKeys.TYPE]
    orch_args = dataline_config[std_dl.DatalineKeys.ORCHESTRATION][
        std_dl.OrchestratorKeys.ARGS]

    flags_to_add = {'project': project_id,
                    'region': gcp_region,
                    'temp_location': pipeline_temp,
                    'staging_location': pipeline_temp}

    api_args = {}
    if orch == 'beam':
        orch_args['setup_file'] = os.path.join(os.getcwd(), 'setup.py')
        orch_args['job_name'] = 'gdp-' + run_name
        orch_args.update(flags_to_add)
        orch_args = parse_yaml_beam_args(orch_args)
    elif orch == 'gcp':
        pass
    elif orch == 'ce':
        orch_args['setup_file'] = os.path.join(os.getcwd(), 'setup.py')
        orch_args['job_name'] = 'gdp-' + run_name
        orch_args.update(flags_to_add)
        orch_args = parse_yaml_beam_args(orch_args)
        api_args = dataline_config[std_env.GlobalKeys.API_ARGS_]
    elif orch == 'kubeflow':
        # TODO: Kubeflow specific stuff here
        pass
    else:
        raise AssertionError('Unknown orchestrator!')

    # Executor
    exec_args = dataline_config[std_dl.DatalineKeys.EXECUTION]
    dist_path = os.path.join(os.getcwd(), 'dist')

    # Need to add empty requirements because of TFX bug in dependency_utils.py
    req_path = os.path.join(dist_path, 'requirements.txt')
    io_utils.write_string_file(req_path, '')
    exec_args['requirements_file'] = req_path

    exec_args['extra_package'] = '/tfx-src/dist/{}-{}.tar.gz'.format(
        PACKAGE_NAME,
        PACKAGE_VERSION)
    exec_args.update(flags_to_add)
    exec_args = parse_yaml_beam_args(exec_args)

    # Metadata Store
    metadata_config = create_metadata_connection_config(
        dataline_config[std_dl.DatalineKeys.METADATA_ARGS],
        pipeline_name,
        output_base_dir)

    return {
        'pipeline_name': pipeline_name,
        'pipeline_root': pipeline_root,
        'pipeline_enable_cache': pipeline_enable_cache,
        'pipeline_log_root': pipeline_log,
        'metadata_connection_config': metadata_config,
        'orchestrator': orch,
        'orchestration_args': orch_args,
        'execution_args': exec_args,
        'datasource': datasource,
        'destination': destination,
        'schema': schema,
        std_env.GlobalKeys.API_ARGS_: api_args,
    }


def get_inferline_spec(inferline_config):
    """
    Args:
        inferline_config:
    """
    std_if.InferlineKeys.key_check(inferline_config)

    bq_args = inferline_config[std_if.InferlineKeys.BQ_ARGS]
    train_args = inferline_config[std_if.InferlineKeys.TRAIN_CONFIG]
    model_uri = inferline_config[std_if.InferlineKeys.MODEL_URI]
    schema_uri = inferline_config[std_if.InferlineKeys.SCHEMA_URI]

    # Pipeline settings
    pipeline_name = inferline_config[std_if.InferlineKeys.EXPERIMENT_NAME]
    pipeline_enable_cache = inferline_config[std_if.InferlineKeys.ENABLE_CACHE]
    project_id = inferline_config[std_if.InferlineKeys.PROJECT_ID]
    gcp_region = inferline_config[std_if.InferlineKeys.GCP_REGION]

    # Pipeline directories
    output_base_dir = inferline_config[std_if.InferlineKeys.OUTPUT_BASE_DIR]
    if not output_base_dir.startswith('gs://'):
        output_base_dir = os.path.abspath(output_base_dir)

    pipeline_root = os.path.join(output_base_dir, pipeline_name)
    run_name = generate_unique_name(pipeline_name)
    pipeline_log = os.path.join(pipeline_root, 'logs', run_name)
    pipeline_temp = os.path.join(pipeline_root, 'tmp', run_name)

    # Orchestrator
    orch_dict = inferline_config[std_if.InferlineKeys.ORCHESTRATION]

    orch = orch_dict[std_dl.OrchestratorKeys.TYPE]
    orch_args = orch_dict[std_dl.OrchestratorKeys.ARGS]

    flags_to_add = {'project': project_id,
                    'region': gcp_region,
                    'temp_location': pipeline_temp,
                    'staging_location': pipeline_temp}

    api_args = {}
    if orch == 'beam':
        orch_args['setup_file'] = os.path.join(os.getcwd(), 'setup.py')
        orch_args['job_name'] = 'gdp-' + run_name
        orch_args.update(flags_to_add)
        orch_args = parse_yaml_beam_args(orch_args)
    elif orch == 'gcp':
        pass
    elif orch == 'ce':
        api_args = inferline_config[std_env.GlobalKeys.API_ARGS_]
        orch_args['setup_file'] = os.path.join(os.getcwd(), 'setup.py')
        orch_args['job_name'] = 'gdp-' + run_name
        orch_args.update(flags_to_add)
        orch_args = parse_yaml_beam_args(orch_args)
    elif orch == 'kubeflow':
        # TODO: Kubeflow specific stuff here
        pass
    else:
        raise AssertionError('Unknown orchestrator!')

    # Executor
    exec_args = inferline_config[std_if.InferlineKeys.EXECUTION]
    exec_args['extra_package'] = '/tfx-src/dist/{}-{}.tar.gz'.format(
        PACKAGE_NAME,
        PACKAGE_VERSION)
    exec_args.update(flags_to_add)
    exec_args = parse_yaml_beam_args(exec_args)

    # Metadata Store
    metadata_config = create_metadata_connection_config(
        inferline_config[std_if.InferlineKeys.METADATA_ARGS],
        pipeline_name,
        output_base_dir)

    return {
        'pipeline_name': pipeline_name,
        'pipeline_root': pipeline_root,
        'pipeline_enable_cache': pipeline_enable_cache,
        'pipeline_log_root': pipeline_log,
        'metadata_connection_config': metadata_config,
        'orchestrator': orch,
        'orchestration_args': orch_args,
        'execution_args': exec_args,
        'bq_args': bq_args,
        'train_config': train_args,
        'model_uri': model_uri,
        'schema_uri': schema_uri,
        std_env.GlobalKeys.API_ARGS_: api_args,
    }



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

import os
import re
from typing import Text, Dict, Any, Tuple

import apache_beam as beam
from google.cloud import storage
from google.oauth2 import service_account as sa_service

from zenml.core.components.data_gen.constants import BINARY_DATA, FILE_NAME, \
    BASE_PATH, FILE_PATH, FILE_EXT
from zenml.core.components.data_gen.constants import GCSImageArgs
from zenml.core.components.data_gen.sources.interface import SourceInterface


# @beam.ptransform_fn
# @beam.typehints.with_input_types(beam.Pipeline)
# @beam.typehints.with_output_types(beam.typehints.Dict[Text, Any])
class ReadGCSBlobs(beam.DoFn):
    def __init__(self, service_account, *unused_args, **unused_kwargs):
        """This is the only way to not do crazy client instantiation cycles.

        Args:
            service_account:
            *unused_args:
            **unused_kwargs:
        """
        super().__init__(*unused_args, **unused_kwargs)
        self._sa = service_account

    def parse_gcs_path(self, gcs_path, object_optional=False):
        """Return the bucket and object names of the given gs:// path.

        Args:
            gcs_path:
            object_optional:
        """
        match = re.match('^gs://([^/]+)/(.*)$', gcs_path)
        if match is None or (match.group(2) == '' and not object_optional):
            raise ValueError(
                'GCS path must be in the form gs://<bucket>/<object>.')
        return match.group(1), match.group(2)

    def start_bundle(self):
        self.credentials = sa_service.Credentials.from_service_account_info(
            self._sa)
        self.gcs_client = storage.Client(credentials=self.credentials)

    def process(self, element, *args, **kwargs):
        """Parse the blob and download as bytestring

        Args:
            element:
            *args:
            **kwargs:
        """
        bucket_name, blob_name = self.parse_gcs_path(element)
        bucket = self.gcs_client.bucket(bucket_name)
        blob = bucket.blob(blob_name)
        blob_as_bstring = blob.download_as_string()
        yield element, blob_as_bstring


@beam.ptransform_fn
@beam.typehints.with_input_types(beam.Pipeline)
@beam.typehints.with_output_types(beam.typehints.Dict[Text, Any])
def ReadFilesFromGCS(pipeline: beam.Pipeline,
                     source_args: Dict[Text, Any]) -> beam.pvalue.PCollection:
    """
    Args:
        pipeline (beam.Pipeline):
        source_args:
    """
    gcs_path = source_args['base_path']
    bucket_re = re.match('^gs://([^/]+)', gcs_path)
    bucket = bucket_re.group(1)
    if bucket_re is None:
        raise ValueError(
            'GCS base path must be in the form gs://<bucket>/.')

    sa = source_args['service_account']

    credentials = sa_service.Credentials.from_service_account_info(sa)
    client = storage.Client(project=sa["project_id"], credentials=credentials)

    blob_list = client.list_blobs(bucket_or_name=bucket)
    uri_list = list(os.path.join(gcs_path, blob.name) for blob in blob_list)

    files_and_contents = (pipeline
                          | beam.Create(uri_list)
                          | beam.ParDo(ReadGCSBlobs(sa))
                          | beam.Reshuffle()
                          | beam.Map(read_file_content, gcs_path)
                          )

    return files_and_contents


def read_file_content(path_and_file: Tuple[Text, bytes], gcs_path: Text):
    """
    Args:
        path_and_file:
        gcs_path (Text):
    """
    data_dict = {FILE_PATH: path_and_file[0],
                 BINARY_DATA: path_and_file[1],
                 BASE_PATH: gcs_path}

    base = os.path.basename(data_dict[FILE_PATH])
    data_dict[FILE_NAME] = os.path.splitext(base)[0]
    data_dict[FILE_EXT] = os.path.splitext(base)[1]

    return data_dict


class GCSImageSource(SourceInterface):
    ARG_KEYS = GCSImageArgs

    def beam_transform(self):
        return ReadFilesFromGCS



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Definition of a the base Backend"""

from typing import Dict

from zenml.utils.print_utils import to_pretty_string, PrintStyles


class BaseBackend:
    """Base class for all ZenML backends.

    Every ZenML pipeline runs in backends
    """
    BACKEND_TYPE = None
    BACKEND_KEY = None

    def __init__(self, **kwargs):
        pass

    def __str__(self):
        return to_pretty_string(self.to_config())

    def __repr__(self):
        return to_pretty_string(self.to_config(), style=PrintStyles.PPRINT)

    def to_config(self):
        """Converts Backend to ZenML config block."""
        return {
            self.BACKEND_KEY: {
                'type': self.BACKEND_TYPE,
                'args': self.__dict__  # everything in init
            }
        }

    @classmethod
    def from_config(cls, config: Dict):
        """
        Convert from ZenML config dict to ZenML Backend object.

        Args:
            config: a ZenML config in dict-form (probably loaded from YAML).
        """
        from zenml.core.backends.backend_factory import backend_factory
        backend_class = backend_factory.get_backend_by_type(config['type'])
        backend_args = config['args']
        obj = backend_class(**backend_args)
        return obj



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

from zenml.core.pipelines.config.standard_config import ConfigKeys


class InferlineKeys(ConfigKeys):
    EXPERIMENT_NAME = 'experiment_name'
    OUTPUT_BASE_DIR = 'output_base_dir'
    BQ_ARGS = 'bq_args'
    PROJECT_ID = 'project_id'
    GCP_REGION = 'gcp_region'
    ENABLE_CACHE = 'enable_cache'
    ORCHESTRATION = 'orchestration'
    EXECUTION = 'execution'
    METADATA_ARGS = 'metadata_args'
    API_ARGS_ = 'api_args'

    TRAIN_CONFIG = 'train_config'
    MODEL_URI = 'model_uri'
    SCHEMA_URI = 'schema_uri'





#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Definition of the GCAIP Training Backend"""

from typing import Text

from zenml.core.backends.training.training_local_backend import \
    TrainingLocalBackend


class TrainingGCAIPBackend(TrainingLocalBackend):
    """Runs TrainerStep on Google Cloud AI Platform.

    Every ZenML pipeline runs in backends
    """
    BACKEND_TYPE = 'gcaip'

    def __init__(
            self,
            scale_tier: Text = 'BASIC',
            runtime_version: Text = '2.2',
            python_version: Text = '3.7',
            max_running_time: Text = '7200s',
            **kwargs):
        self.scale_tier = scale_tier
        self.runtime_version = runtime_version
        self.python_version = python_version
        self.max_running_time = max_running_time
        super().__init__(**kwargs)



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Base Step Interface definition"""

from typing import Dict, Type

from zenml.utils.print_utils import to_pretty_string, PrintStyles
from zenml.utils.source_utils import resolve_source_path, \
    load_source_path_class


class BaseStep:
    """Base class for all ZenML steps.

    These are 'windows' into the base components for simpler overrides.
    """

    def __str__(self):
        return to_pretty_string(self.to_config())

    def __repr__(self):
        return to_pretty_string(self.to_config(), style=PrintStyles.PPRINT)

    @staticmethod
    def from_config(config_block: Dict) -> Type:
        """
        Takes config block that represents a Step and converts it back into
        its Python equivalent. This functionality is similar for most steps,
        and expected config_block may look like

        {
            'source': this.module.StepClass@sha  # where sha is optional
            'args': {}  # to be passed to the constructor
        }

        Args:
            config_block: config block representing source and args of step.
        """
        # resolve source path
        if 'source' in config_block:
            source = config_block['source']
            class_ = load_source_path_class(source)
            args = config_block['args']
            return class_(**args)
        else:
            raise AssertionError("Cannot create config_block without source "
                                 "key.")

    def to_config(self) -> Dict:
        """
        Converts from step back to config. This functionality for most steps
        yields the following config block.

        {
            'source': this.file.path.BaseStep@sha  # where sha is optional
            'args': {}  # whatever is used in the constructor for BaseStep
        }
        """
        # resolve git
        source_path = resolve_source_path(
            self.__class__.__module__ + '.' + self.__class__.__name__
        )

        return {
            'source': source_path,
            'args': self.__dict__  # everything in init
        }



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Implementation of the ratio-based categorical split."""

from typing import Text, List, Dict

from zenml.core.steps.split.utils import partition_cat_list
from zenml.core.steps.split import constants
from zenml.core.steps.split.base_split_step import BaseSplitStep
from zenml.core.steps.split.categorical_domain_split_step import \
    CategoricalPartitionFn


def lint_split_map(split_map: Dict[Text, float]):
    """Small utility to lint the split_map"""
    if constants.TRAIN not in split_map.keys():
        raise AssertionError(f'You have to define some values for '
                             f'the {constants.TRAIN} split.')
    if len(split_map) <= 1:
        raise AssertionError('Please specify more than 1 split name in the '
                             'split_map!')


class CategoricalRatioSplitStep(BaseSplitStep):

    def __init__(
            self,
            categorical_column: Text,
            categories: List[Text],
            split_ratio: Dict[Text, float],
            statistics=None,
            schema=None,
    ):
        """

        Args:
            statistics: Parsed statistics from a preceding StatisticsGen.
            schema: Parsed schema from a preceding SchemaGen.
            categorical_column: Name of the categorical column used for
                                splitting.
            split_ratio: A dict { split_name: percentage of categories
                                    in split }.
        """
        self.categorical_column = categorical_column

        split_map = partition_cat_list(categories, split_ratio)
        lint_split_map(split_map)
        self.split_map = split_map

        super().__init__(statistics, schema)

    def partition_fn(self):
        return CategoricalPartitionFn, {
            'split_map': self.split_map,
            'categorical_column': self.categorical_column
        }

    def get_split_names(self) -> List[Text]:
        return list(self.split_map.keys())



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Base Class for all ZenML datasources"""

import os
from typing import Text, Dict
from uuid import uuid4

import tensorflow as tf
from zenml.utils.print_utils import to_pretty_string, PrintStyles
from zenml.core.repo.repo import Repository
from zenml.core.standards import standard_config as keys
from zenml.utils import path_utils
from zenml.utils import source_utils
from zenml.utils.enums import GDPComponent
from zenml.utils.post_training.post_training_utils import \
    get_schema_artifact, view_schema, get_feature_spec_from_schema, \
    convert_raw_dataset_to_pandas
from zenml.utils.zenml_analytics import track, CREATE_DATASOURCE


class BaseDatasource:
    """Base class for all ZenML datasources.

    Every ZenML datasource should override this class.
    """
    DATA_STEP = None
    PREFIX = 'pipeline_'

    @track(event=CREATE_DATASOURCE)
    def __init__(self, name: Text, schema: Dict = None, _id: Text = None,
                 *args, **kwargs):
        """
        Construct the datasource

        Args:
            name (str): name of datasource
            schema (dict): schema of datasource
            _id: unique ID (for internal use)
        """
        if _id:
            self._id = _id
        else:
            # If none, then this is assumed to be 'new'. Check dupes.
            repo: Repository = Repository.get_instance()
            all_names = repo.get_datasource_names()
            if any(d == name for d in all_names):
                raise Exception(
                    f'Datasource {name} already exists! Please '
                    f'use Repository.get_datasource_by_name({name}) to fetch.')
            self._id = str(uuid4())

        self.name = name
        self.schema = schema

    def __str__(self):
        return to_pretty_string(self.to_config())

    def __repr__(self):
        return to_pretty_string(self.to_config(), style=PrintStyles.PPRINT)

    @classmethod
    def get_name_from_pipeline_name(cls, pipeline_name: Text):
        return pipeline_name[len(cls.PREFIX)]

    @classmethod
    def from_config(cls, config: Dict):
        """
        Convert from Data Step config to ZenML Datasource object.

        Data step is also populated and configuration set to parameters set
        in the config file.

        Args:
            config: a DataStep config in dict-form (probably loaded from YAML).
        """
        if keys.DataSteps.DATA not in config[keys.GlobalKeys.STEPS]:
            raise Exception("Cant have datasource without data step.")

        # this is the data step config block
        step_config = config[keys.GlobalKeys.STEPS][keys.DataSteps.DATA]
        datasource_class = source_utils.load_source_path_class(
            config[keys.GlobalKeys.DATASOURCE][keys.DatasourceKeys.SOURCE])
        datasource_name = config[keys.GlobalKeys.DATASOURCE][
            keys.DatasourceKeys.NAME]
        _id = config[keys.GlobalKeys.DATASOURCE][keys.DatasourceKeys.ID]
        return datasource_class(
            name=datasource_name, _id=_id, **step_config['args'])

    def to_config(self):
        """Converts datasource to ZenML config block."""
        return {
            keys.DatasourceKeys.NAME: self.name,
            keys.DatasourceKeys.SOURCE: source_utils.resolve_source_path(
                self.__class__.__module__ + '.' + self.__class__.__name__
            ),
            keys.DatasourceKeys.ID: self._id
        }

    def get_pipeline_name_from_name(self):
        return self.PREFIX + self.name

    def get_data_step(self):
        params = self.__dict__.copy()
        # TODO: [HIGH] Figure out if there is a better way to do this
        params.pop('name')
        params.pop('_id')
        return self.DATA_STEP(**params)

    def _get_one_pipeline(self):
        """Gets representative pipeline from all pipelines associated."""
        repo: Repository = Repository.get_instance()
        pipelines = repo.get_pipelines_by_datasource(self)

        if len(pipelines) == 0:
            raise Exception('This datasource is not associated with any '
                            'pipelines, therefore there is no data!')
        return pipelines[0]

    def _get_data_file_paths(self, pipeline):
        """
        Gets path where data is stored as list of file paths.

        Args:
            pipeline: a pipeline with this datasource embedded
        """
        if pipeline.datasource._id != self._id:
            raise AssertionError('This pipeline does not belong to this '
                                 'datasource.')
        repo: Repository = Repository.get_instance()
        # Take any pipeline and get the datagen
        data_uri = os.path.join(repo.get_artifacts_uri_by_component(
            pipeline.pipeline_name,
            GDPComponent.DataGen.name
        )[0], 'examples')
        data_files = path_utils.list_dir(data_uri)
        return data_files

    def sample_data(self, sample_size: int = 100000):
        """
        Sampels data from datasource as a pandas DataFrame.

        Args:
            sample_size: # of rows to sample.
        """
        pipeline = self._get_one_pipeline()
        data_files = self._get_data_file_paths(pipeline)

        spec = get_feature_spec_from_schema(
            pipeline.pipeline_name, GDPComponent.DataSchema.name)

        dataset = tf.data.TFRecordDataset(data_files, compression_type='GZIP')
        return convert_raw_dataset_to_pandas(dataset, spec, sample_size)

    def get_datapoints(self):
        """Gets total number of datapoints in datasource"""
        # TODO: [HIGH] Refactor this
        pipeline = self._get_one_pipeline()
        data_files = self._get_data_file_paths(pipeline)
        return sum(1 for _ in tf.data.TFRecordDataset(data_files,
                                                      compression_type='GZIP'))

    def view_schema(self):
        """View schema of data flowing in pipeline."""
        pipeline = self._get_one_pipeline()
        uri = get_schema_artifact(
            pipeline.pipeline_name, GDPComponent.DataSchema.name)
        view_schema(uri)



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Factory to register step classes to steps"""

from zenml.core.steps.deployer import DeployerStep
from zenml.core.steps.evaluator import EvaluatorStep
from zenml.core.steps.preprocesser import PreprocesserStep
from zenml.core.steps.split.base_split_step import BaseSplitStep
from zenml.core.steps.timeseries import TimeseriesStep
from zenml.core.steps.trainer import TrainerStep


class StepFactory:
    """Definition of StepFactory to track all steps in ZenML.

    All steps (including custom steps) are to be registered here.
    """

    def __init__(self):
        self.steps = {}

    def get_steps(self):
        return self.steps

    def get_single_step(self, key):
        return self.steps[key]

    def register_step(self, key, step_):
        self.steps[key] = step_


# Register the injections into the factory
type_factory = StepFactory()
type_factory.register_step('trainer', TrainerStep)
type_factory.register_step('split', BaseSplitStep)
type_factory.register_step('evaluator', EvaluatorStep)
type_factory.register_step('timeseries', TimeseriesStep)
type_factory.register_step('preprocessor', PreprocesserStep)
type_factory.register_step('deployer', DeployerStep)



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Base interface for Image Data Step"""

import json
import os
from typing import Dict, Text, Any

import apache_beam as beam
from apache_beam.io import fileio

from zenml.core.components.data_gen.constants import BINARY_DATA, FILE_NAME, \
    FILE_EXT, METADATA, LABEL
from zenml.core.steps.data.base_data_step import BaseDataStep


@beam.ptransform_fn
@beam.typehints.with_input_types(beam.Pipeline)
@beam.typehints.with_output_types(beam.typehints.Dict[Text, Any])
def ReadImagesFromDisk(pipeline: beam.Pipeline,
                       base_path: Text) -> beam.pvalue.PCollection:
    """
    Args:
        pipeline (beam.Pipeline): Pipeline object built by the DataGen
        executor.
        base_path (Text): Base directory containing images and labels.
    """

    wildcard_qualifier = "**"

    # ingest all the files from the base path by supplying the wildcard
    file_pattern = os.path.join(base_path, wildcard_qualifier)

    allowed_ext = [".jpg", ".json", ".png", ".txt", ".jpeg"]

    images, label_file = (pipeline
                          | fileio.MatchFiles(file_pattern)
                          | fileio.ReadMatches()
                          | beam.Map(read_file_content)
                          | "FilterOutFiles" >> beam.Filter(lambda x:
                                                            x[FILE_EXT] in
                                                            allowed_ext)
                          | "SplitLabelFile" >> beam.Partition(SplitByFileName,
                                                               2))

    # label_file is actually a dict
    label_dict = beam.pvalue.AsSingleton(label_file)
    ready_images = (images
                    | "AddLabelAndMetadata" >> beam.Map(add_label_and_metadata,
                                                        label_dict))

    return ready_images


class ImageDataStep(BaseDataStep):
    def __init(self, base_path, schema: Dict = None):
        super(ImageDataStep, self).__init__(schema=schema)
        self.base_path = base_path

    def read_from_source(self):
        return ReadImagesFromDisk(self.base_path)


def read_file_content(file: beam.io.fileio.ReadableFile):
    """
    Args:
        file (beam.io.fileio.ReadableFile): Beam ReadableFile object,
        corresponds to an image file read from disk.
    """
    file_path = file.metadata.path
    base = os.path.basename(file_path)

    data_dict = {BINARY_DATA: file.read(),
                 FILE_NAME: base,
                 FILE_EXT: os.path.splitext(base)[1]}

    return data_dict


@beam.typehints.with_input_types(Dict[Text, Any], int)
@beam.typehints.with_output_types(int)
def SplitByFileName(element: Dict[Text, Any],
                    num_partitions: int) -> int:
    """
    Args:
        element: Dict with image features.
        num_partitions (int): Number of partitions, unused.
    """
    filename = element[FILE_NAME]
    file_ext = element[FILE_EXT]

    if "label" in filename:
        return 1

    if file_ext.lower() in [".txt", ".json", ".csv"]:
        return 1

    return 0


def add_label_and_metadata(image_dict: Dict[Text, Any], label_dict: Text):
    """Add label and metadata information to an image.

    Args:
        image_dict: Dict with image features.
        label_dict (Text): JSON-readable string with label information.
    """
    filename = image_dict[FILE_NAME]

    label_data = label_dict[BINARY_DATA]
    label, metadata = get_matching_label(label_data, filename)

    image_dict.update(metadata)
    image_dict[LABEL] = label

    return image_dict


def get_matching_label(label_data: Text, img_filename: Text):
    """ Get a label matching an image file name from a
        JSON-readable label file.
    Args:
        label_data (Text): Label string, needs to be JSON-readable.
        img_filename (Text): File name of the image.

    Returns:
        * **label** (*Str, label key*)
        * **metadata** (*Dict, additional metadata information*)
    """

    # This can potentially go bad if the JSON is too big
    data_dict = json.loads(label_data)

    img_data = data_dict[img_filename]

    if "label" not in img_data:
        raise ValueError("Error while parsing label file: No label "
                         "key was found. Make sure to specify your "
                         "image label under a \"label\" key.")

    label = img_data.pop(LABEL)

    try:
        metadata = img_data.pop(METADATA)
    except KeyError:
        metadata = {}

    # JSON parsing puts empty metadata dict as the empty string,
    # so in that case coerce to dict just to be type safe
    if not metadata:
        metadata = {}

    return label, metadata



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

import abc
from typing import Optional, Text

from tfx.components.base import base_component
from tfx.types import channel_utils, artifact_utils, standard_artifacts


class SessionTransform(base_component.BaseComponent):
    """Custom SessionTransform component

    This component is created as a baseline to apply transformations on a
    session level .

    In order to use the baseline for creating a component, the user needs to
    follow three steps:

        1 - Create a new executor class inheriting from the SessionExecutor
            and overwrite the method called 'get_combiner' which needs to return
            the CombineFn which processes sessions

        2 - Create a new component class inheriting from the SessionComponent
            and set the EXECUTOR_SPEC using the executor mentioned in step 1

        3 - Implement the config parser for the specific use-case
    """

    SPEC_CLASS = None

    def __init__(self,
                 examples=None,
                 schema=None,
                 config=None,
                 instance_name: Optional[Text] = None):
        """
        Args:
            examples:
            schema:
            config:
            instance_name:
        """
        self.exp_dict = config
        self.schema = schema
        config_dict = self.config_parser(config)

        output_artifact = standard_artifacts.Examples()
        output_artifact.split_names = artifact_utils.encode_split_names(
            ['train', 'eval'])

        output = channel_utils.as_channel([output_artifact])

        spec = self.SPEC_CLASS(
            input=examples,
            schema=schema,
            output=output,
            **config_dict
        )
        super(SessionTransform, self).__init__(spec=spec,
                                               instance_name=instance_name)

    @abc.abstractmethod
    def config_parser(self):
        """Abstract method which needs to be implemented to parse the relevant
        configuration out of the experiment config file
        """
        return {}



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Base ZenML repository"""

import os
from pathlib import Path
from typing import Text, List, Dict, Any, Optional

from zenml.core.metadata.metadata_wrapper import ZenMLMetadataStore
from zenml.core.repo.artifact_store import ArtifactStore
from zenml.core.repo.git_wrapper import GitWrapper
from zenml.core.repo.global_config import GlobalConfig
from zenml.core.repo.zenml_config import ZenMLConfig
from zenml.core.standards import standard_config as keys
from zenml.utils import path_utils, yaml_utils
from zenml.utils.enums import PipelineStatusTypes
from zenml.utils.zenml_analytics import track, CREATE_REPO, GET_PIPELINES, \
    GET_DATASOURCES, GET_PIPELINE_ARTIFACTS, GET_STEPS, REGISTER_PIPELINE


class Repository:
    """ZenML repository definition. This is a Singleton class.

    Every ZenML project exists inside a ZenML repository.
    """
    __instance__ = None

    def __init__(self, path: Text = None):
        """
        Construct reference a ZenML repository

        Args:
            path (str): Path to root of repository
        """
        if Repository.__instance__ is None:
            if path is None:
                # Start from cwd and traverse up until we find zenml config.
                path = Repository.get_zenml_dir(os.getcwd())

            if not path_utils.is_dir(path):
                raise Exception(f'{path} does not exist or is not a dir!')
            self.path = path

            # Hook up git
            self.git_wrapper = GitWrapper(self.path)

            # Load the ZenML config
            self.zenml_config = ZenMLConfig(self.path)

            Repository.__instance__ = self
        else:
            raise Exception("You cannot create another Repository class!")

    @staticmethod
    def get_zenml_dir(path: Text):
        """
        Recursive function to find the zenml config starting from path.

        Args:
            path (str): starting path
        """
        if ZenMLConfig.is_zenml_dir(path):
            return path

        if path_utils.is_root(path):
            raise Exception(
                'Looks like you used ZenML outside of a ZenML repo. '
                'Please init a ZenML repo first before you using '
                'the framework.')
        return Repository.get_zenml_dir(Path(path).parent)

    @staticmethod
    def get_instance():
        """ Static method to fetch the current instance."""
        if not Repository.__instance__:
            Repository()
        return Repository.__instance__

    @staticmethod
    @track(event=CREATE_REPO)
    def init_repo(repo_path: Text, artifact_store_path: Text = None,
                  metadata_store: Optional[ZenMLMetadataStore] = None,
                  pipelines_dir: Text = None,
                  analytics_opt_in: bool = None):
        """
        Initializes a git repo with zenml.

        Args:
            repo_path (str): path to root of a git repo
            metadata_store: metadata store definition
            artifact_store_path (str): path where to store artifacts
            pipelines_dir (str): path where to store pipeline configs.
            analytics_opt_in (str): opt-in flag for analytics code.

        Raises:
            InvalidGitRepositoryError: If repository is not a git repository.
            NoSuchPathError: If the repo_path does not exist.
        """
        # check whether its a git repo by initializing GitWrapper
        GitWrapper(repo_path)

        # use the underlying ZenMLConfig class to create the config
        ZenMLConfig.create_config(
            repo_path, artifact_store_path, metadata_store, pipelines_dir)

        # create global config
        global_config = GlobalConfig.get_instance()
        if analytics_opt_in is not None:
            global_config.set_analytics_opt_in(analytics_opt_in)

    def get_artifact_store(self) -> ArtifactStore:
        return self.zenml_config.get_artifact_store()

    def get_metadata_store(self):
        return self.zenml_config.get_metadata_store()

    def get_git_wrapper(self) -> GitWrapper:
        return self.git_wrapper

    @track(event=GET_STEPS)
    def get_steps(self):
        """List all registered steps in repository"""
        pass

    def get_datasource_by_name(self, name: Text) -> List:
        """
        Get all datasources in this repo.

        Returns: list of datasources used in this repo
        """
        all_datasources = self.get_datasources()
        for d in all_datasources:
            if name == d.name:
                return d
        raise Exception(f'Datasource {name} does not exist')

    @track(event=GET_DATASOURCES)
    def get_datasources(self) -> List:
        """
        Get all datasources in this repo.

        Returns: list of datasources used in this repo
        """
        from zenml.core.datasources.base_datasource import BaseDatasource

        datasources = []
        for file_path in self.get_pipeline_file_paths():
            c = yaml_utils.read_yaml(file_path)
            datasources.append(BaseDatasource.from_config(c))
        return datasources

    def get_datasource_names(self) -> List:
        """
        Get all datasources in this repo.

        Returns: List of datasource names used in this repo.
        """
        n = []
        for file_path in self.get_pipeline_file_paths():
            c = yaml_utils.read_yaml(file_path)
            n.append(c[keys.GlobalKeys.DATASOURCE][keys.DatasourceKeys.NAME])
        return list(set(n))

    @track(event=GET_PIPELINES)
    def get_pipelines(self, type_filter: List[Text] = None) -> List:
        """
        Gets list of pipelines, optionally filtered by type.

        Args:
            type_filter (list): list of types to filter by.
        """
        from zenml.core.pipelines.base_pipeline import BasePipeline
        pipelines = []
        for file_path in self.get_pipeline_file_paths():
            c = yaml_utils.read_yaml(file_path)
            pipelines.append(BasePipeline.from_config(c))

        if type_filter:
            # filter by type
            return [p for p in pipelines if p.pipeline_type in type_filter]
        else:
            # return all
            return pipelines

    def get_pipeline_names(self) -> Optional[List[Text]]:
        """Gets list of pipeline (unique) names"""
        store = self.zenml_config.get_metadata_store()

        contexts = store.get_pipeline_contexts()
        return [c.name for c in contexts]

    def get_pipeline_file_paths(self, only_file_names: bool = False) -> \
            Optional[List[Text]]:
        """Gets list of pipeline file path"""
        pipelines_dir = self.zenml_config.get_pipelines_dir()

        if not path_utils.is_dir(pipelines_dir):
            return []
        return path_utils.list_dir(pipelines_dir, only_file_names)

    @track(event=GET_PIPELINES)
    def get_pipelines_by_datasource(self, datasource):
        """
        Gets list of pipelines associated with datasource.

        Args:
            datasource (BaseDatasource): object of type BaseDatasource.
        """
        from zenml.core.pipelines.base_pipeline import BasePipeline
        pipelines = []
        for file_path in self.get_pipeline_file_paths():
            c = yaml_utils.read_yaml(file_path)
            if c[keys.GlobalKeys.DATASOURCE][keys.DatasourceKeys.ID] == \
                    datasource._id:
                pipelines.append(BasePipeline.from_config(c))
        return pipelines

    def get_hyperparameters(self):
        """
        Hyper-parameter list of all pipelines in repo
        """
        pass

    @track(event=GET_PIPELINE_ARTIFACTS)
    def get_artifacts_uri_by_component(self, pipeline_name: Text,
                                       component_name: Text):
        """
        Gets the artifacts of any component within a pipeline. All artifacts
        are resolved locally, even if artifact store is remote.

        Args:
            pipeline_name (str): name of pipeline
            component_name (str): name of component
        """
        # mlmd
        metadata_store = self.get_metadata_store()
        status = metadata_store.get_pipeline_status(pipeline_name)
        if status != PipelineStatusTypes.Succeeded.name:
            AssertionError('Cannot retrieve as pipeline is not succeeded.')
        # if component_name not in GDPCOmponents:
        #     raise AssertionError('Component must be one of {}')
        artifacts = metadata_store.get_artifacts_by_component(pipeline_name,
                                                              component_name)

        # Download if not local
        uris = []
        for a in artifacts:
            artifact_store = self.get_artifact_store()
            uris.append(artifact_store.resolve_uri_locally(a.uri))

        return uris

    def get_hyperparameters_pipeline(self):
        pass

    @track(event=REGISTER_PIPELINE)
    def register_pipeline(self, file_name: Text, config: Dict[Text, Any]):
        """
        Registers a pipeline in the artifact store as a YAML file.

        Args:
            file_name (dict): file name of pipeline
            config (dict): dict representation of ZenML config
        """
        pipelines_dir = self.zenml_config.get_pipelines_dir()

        # Create dir
        path_utils.create_dir_if_not_exists(pipelines_dir)

        # Write
        yaml_utils.write_yaml(os.path.join(pipelines_dir, file_name), config)

    def compare_pipelines(self):
        pass



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Wrapper class to handle Git integration"""

import logging
import os
from typing import Text, List

from git import Repo as GitRepo

from zenml.core.repo.constants import GIT_FOLDER_NAME
from zenml.utils import path_utils


class GitWrapper:
    """Wrapper class for Git.

    This class is responsible for handling git interactions, primarily
    handling versioning of different steps in pipelines.
    """

    def __init__(self, repo_path: Text):
        """
        Initialize GitWrapper. Should be initialize by ZenML Repository.
        Args:
            repo_path:

        Raises:
            InvalidGitRepositoryError: If repository is not a git repository.
            NoSuchPathError: If the repo_path does not exist.
        """
        # TODO: [LOW] Raise ZenML exceptions here instead.
        self.repo_path: Text = repo_path
        self.git_root_path: Text = os.path.join(repo_path, GIT_FOLDER_NAME)
        self.git_repo: GitRepo = GitRepo(self.repo_path)

        # TODO: [LOW] Check whether this is necessary.
        assert not self.git_repo.bare

    def add_gitignore(self, items: List[Text]):
        """
        Adds `items` to .gitignore, if .gitignore exists. Otherwise creates
        and adds.

        Args:
            items (list[str]): Items to add.
        """
        str_items = '\n'.join(items)
        str_items = '\n\n# ZenML\n' + str_items

        gitignore_path = os.path.join(self.repo_path, '.gitignore')
        if not path_utils.file_exists(gitignore_path):
            path_utils.create_file_if_not_exists(gitignore_path, str_items)
        else:
            path_utils.append_file(gitignore_path, str_items)

    def check_file_committed(self, file_path: Text) -> bool:
        """
        Checks file is committed. If yes, return True, else False.

        Args:
            file_path (str): Path to any file within the ZenML repo.
        """
        for item in self.git_repo.index.diff(None):
            # These are all changed files
            if file_path == item.a_path:
                return False
        return True

    def get_current_sha(self) -> Text:
        """
        Finds the git sha that each file within the module is currently on.

        Args:
            module_path (str): Path to any module within the ZenML repo.
        """
        # TODO: [HIGH] Check whether returning the head makes sense.
        return self.git_repo.head.object.hexsha

    def resolve_source_path(self, source_path: Text) -> Text:
        """
        Takes source path (e.g. this.module.ClassName), and appends relevant
        sha to it if the files within `module` are all committed. If even one
        file is not committed, then returns `source_path` unchanged.

        Args:
            source_path (str): relative module path pointing to a Class.
        """
        # import here to resolve circular dependency
        from zenml.utils.source_utils import get_path_from_source
        file_path = get_path_from_source(source_path)

        if not self.check_file_committed(file_path):
            logging.warning(f'Found uncommitted file {file_path}. '
                            f'Pipelines run with this configuration '
                            f'may not be reproducible. Please commit all '
                            f'files '
                            f'in this module and then run the pipeline to '
                            f'ensure reproducibility.')
            return source_path
        return source_path + '@' + self.get_current_sha()

    def stash(self):
        """Wrapper for git stash"""
        git = self.git_repo.git
        git.stash()

    def stash_pop(self):
        """Wrapper for git stash pop. Only pops if there's something to pop."""
        git = self.git_repo.git
        if git.stash('list') != '':
            git.stash('pop')

    def checkout(self, sha_or_branch_name: Text):
        """
        Wrapper for git checkout

        Args:
            sha_or_branch_name: hex string of len 40 representing git sha OR
            name of branch
        """
        # TODO: [HIGH] Implement exception handling
        git = self.git_repo.git
        git.checkout(sha_or_branch_name)



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Implementation of a random split of the input data set."""

import bisect
from typing import Text, List, Dict, Any

import numpy as np

from zenml.core.steps.split import constants
from zenml.core.steps.split.base_split_step import BaseSplitStep


def lint_split_map(split_map: Dict[Text, float]):
    """Small utility to lint the split_map"""
    if constants.TRAIN not in split_map.keys():
        raise AssertionError(f'You have to define some values for '
                             f'the {constants.TRAIN} split.')
    if len(split_map) <= 1:
        raise AssertionError('Please specify more than 1 split name in the '
                             'split_map!')

    if not all(isinstance(v, float) for v in split_map.values()):
        raise AssertionError("Only float values are allowed when specifying "
                             "a random split!")


def RandomSplitPartitionFn(element: Any,
                           num_partitions: int,
                           split_map: Dict[Text, float]) -> int:
    """
    Function for a random split of the data; to be used in a beam.Partition.
    Args:
        element: Data point, in format tf.train.Example.
        num_partitions: Number of splits, unused here.
        split_map: Dict mapping {split_name: percentage of data in split}.

    Returns: An integer n, where 0 <= n <= num_partitions - 1.
    """

    # calculates probability mass of each split (= interval on [0,1))
    probability_mass = np.cumsum(list(split_map.values()))

    # excludes the right bound 1 as it is redundant
    brackets = probability_mass[:-1]

    return bisect.bisect(brackets, np.random.rand())


class RandomSplitStep(BaseSplitStep):

    def __init__(
            self,
            split_map: Dict[Text, float],
            statistics=None,
            schema=None,
    ):
        """

        Args:
            statistics: Parsed statistics from a preceding StatisticsGen.
            schema: Parsed schema from a preceding SchemaGen.
            split_map: A dict { split_name: percentage of data in split }.
        """

        lint_split_map(split_map)
        self.split_map = split_map

        super().__init__(statistics, schema)

    def partition_fn(self):
        return RandomSplitPartitionFn, {
            'split_map': self.split_map
        }

    def get_split_names(self) -> List[Text]:
        return list(self.split_map.keys())



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

import json
import os

import tensorflow as tf
from zenml.utils.standard_formats import UDFStandard

from zenml.utils.component_utils import parse_methods, get_bq_schema
from zenml.utils.component_utils import timer
from zenml.utils.constants import ZENML_PIPELINE_CONFIG
from zenml.utils.standards.standard_config import TransformMethods, \
    GlobalKeys, PreProcessKeys, PCAKeys, NonSeqFillingMethods, MethodKeys

PCA_PREFIX = 'pca_'


def apply_filling(data, description):
    """
    Args:
        data:
        description:
    """
    method_name = description[MethodKeys.METHOD]
    method_params = description[MethodKeys.PARAMETERS]
    return NonSeqFillingMethods.get_method(method_name)(**method_params)(data)


@timer
def preprocessing_fn(inputs):
    """
    Args:
        inputs:
    """
    config = json.loads(os.getenv(ZENML_PIPELINE_CONFIG))

    pca_config = config[GlobalKeys.PCA_]
    PCAKeys.key_check(pca_config)

    num_dims = pca_config[PCAKeys.NUM_DIMENSIONS]
    bq_schema = get_bq_schema(
        config[GlobalKeys.BQ_ARGS_],
        experiment_dict=config,
    )

    filling_dict = dict()
    for key in [GlobalKeys.FEATURES]:
        filling_dict.update(parse_methods(config[key],
                                          config[GlobalKeys.PREPROCESSING],
                                          PreProcessKeys.FILLING,
                                          bq_schema,
                                          NonSeqFillingMethods))

    # calculate normalized features
    norm_vecs = []
    pca_output = {}
    for key, value in inputs.items():
        if key in config[GlobalKeys.FEATURES] and bq_schema[key] != 'STRING':
            # TODO: [High] Add a way of imputing missing values
            value = apply_filling(value, filling_dict[key])

            # normalize to z score
            normed = TransformMethods.get_method('scale_to_z_score')(value)

            # cast of float32
            floated = tf.cast(normed, dtype=tf.float32)
            norm_vecs.append(floated)
        pca_output[key] = value  # also store the raw feature

    # concatenate all the norm features and calculate eigenvectors
    # multiply eigenvector matrix by features to get reduced features
    feature_matrix = tf.concat(norm_vecs, axis=1)
    pca = TransformMethods.get_method('pca')(feature_matrix,
                                             output_dim=num_dims,
                                             dtype=tf.float32)
    pca_features = tf.linalg.matmul(feature_matrix, pca)
    pca_features = tf.unstack(pca_features, axis=1)
    for i, f in enumerate(pca_features):
        pca_output[UDFStandard.ANALYZE_PREFIX + PCA_PREFIX + str(i)] = f

    return pca_output



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

from zenml.core.components.data_gen import constants
# from zenml.core.components.data_gen.sources.bq import BigQuerySource
from zenml.core.components.data_gen.sources.local_image import LocalImageSource
from zenml.core.components.data_gen.sources.gcs_image import GCSImageSource
from zenml.core.components.data_gen.sources.local_csv import LocalCSVSource
# from zenml.core.components.data_gen.sources.gcs_csv import GCSCSVSource


class SourceFactory:

    def __init__(self):
        self.sources = {}

    def get_sources(self):
        return self.sources

    def get_single_source(self, key):
        """
        Args:
            key:
        """
        return self.sources[key]

    def register_source(self, key, source):
        """
        Args:
            key:
            source:
        """
        self.sources[key] = source


# Register the injections into the factory
source_factory = SourceFactory()
# source_factory.register_source(constants.SOURCE_BQ, BigQuerySource)
source_factory.register_source(constants.SOURCE_LOCAL_IMAGE, LocalImageSource)
source_factory.register_source(constants.SOURCE_GCS_IMAGE, GCSImageSource)
source_factory.register_source(constants.SOURCE_LOCAL_CSV, LocalCSVSource)
# source_factory.register_source(constants.SOURCE_GCS_CSV, GCSCSVSource)



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

import abc

from zenml.core.components.data_gen.constants import DatasourceKeys


class SourceInterface:
    ARG_KEYS = None

    @classmethod
    def config_parser(cls, config):
        """
        Args:
            config:
        """
        DatasourceKeys.key_check(config)
        args = config[DatasourceKeys.ARGS]
        cls.ARG_KEYS.key_check(args)
        return config

    @abc.abstractmethod
    def beam_transform(self, *args, **kwargs):
        """
        Args:
            *args:
            **kwargs:
        """
        pass



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

from abc import abstractmethod
from typing import Text, List, Dict

import pandas as pd
from ml_metadata.metadata_store import metadata_store
from ml_metadata.proto import metadata_store_pb2

from zenml.utils.enums import PipelineStatusTypes
from zenml.utils.logger import get_logger
from zenml.utils.print_utils import to_pretty_string, PrintStyles

logger = get_logger(__name__)


class ZenMLMetadataStore:
    STORE_TYPE = None

    def __init__(self, **kwargs):
        if 'store' in kwargs:
            self.store = kwargs['store']
        else:
            self.store = None

    @staticmethod
    def initiate_store_mysql(
            host: Text,
            port: int,
            database: Text,
            username: Text,
            password: Text
    ) -> metadata_store_pb2.ConnectionConfig:
        """Initiates the configuration of the store. :param host: The name or
        network address of the instance of MySQL to :param connect to.: :param
        port: The port MySQL is using to listen for connections. :param
        database: The name of the database to use. :param username: The MySQL
        login account being used. :param password: The password for the MySQL
        account being used.

        Args:
            host (Text):
            port (int):
            database (Text):
            username (Text):
            password (Text):

        Returns:
            the instance of the MD Store
        """
        config = metadata_store_pb2.ConnectionConfig(
            mysql=metadata_store_pb2.MySQLDatabaseConfig(
                host=host,
                port=port,
                database=database,
                user=username,
                password=password))

        return metadata_store.MetadataStore(config)

    def __str__(self):
        return to_pretty_string(self.to_config())

    def __repr__(self):
        return to_pretty_string(self.to_config(), style=PrintStyles.PPRINT)

    @staticmethod
    def initiate_store_sqlite(uri: Text):
        """Initiates the configuration of the store :param uri: path the the
        metadata store :return: the instance of the MD Store

        Args:
            uri (Text):
        """
        config = metadata_store_pb2.ConnectionConfig()
        config.sqlite.filename_uri = uri
        return metadata_store.MetadataStore(config)

    @classmethod
    def from_config(cls, config: Dict):
        """
        Converts ZenML config to ZenML Metadata Store.

        Args:
            config (dict): ZenML config block for metadata.
        """
        from zenml.core.metadata.metadata_wrapper_factory import \
            wrapper_factory
        from zenml.utils.enums import MLMetadataTypes
        store_type = config['type']
        store_types = list(MLMetadataTypes.__members__.keys())
        if store_type not in store_types:
            raise Exception(f'store_type must be one of: {store_types}')
        args = config['args']
        class_ = wrapper_factory.get_single_metadata_wrapper(store_type)

        try:
            obj = class_(**args)
        except TypeError as e:
            logger.error(str(e))
            import inspect
            args = inspect.getfullargspec(class_).args
            args.remove('self')
            raise Exception(f'args must include only: {args}')
        return obj

    @abstractmethod
    def get_tfx_metadata_config(self):
        pass

    def to_config(self) -> Dict:
        """
        Converts from ZenML Metadata store back to config.
        """
        args_dict = self.__dict__.copy()
        args_dict.pop('store')  # TODO: [LOW] Again a hack
        return {
            'type': self.STORE_TYPE,
            'args': args_dict
        }

    def get_pipeline_status(self, pipeline_name: Text) -> Text:
        """
        Query metadata store to find status of pipeline.

        Args:
            pipeline_name (str): name of pipeline
        """
        components_status = self.get_components_status(pipeline_name)
        for status in components_status:
            if status != 'complete' or status != 'cached':
                return PipelineStatusTypes.Running.name
        return PipelineStatusTypes.Succeeded.name

    def get_components_status(self, pipeline_name: Text):
        pipeline_executions = self.get_pipeline_executions(pipeline_name)
        return {
            e.properties['component_id'].string_value: e.properties[
                'state'].string_value for e in
            pipeline_executions
        }

    def get_pipeline_executions(self, pipeline_name: Text):
        c = self.get_pipeline_context(pipeline_name)
        return self.store.get_executions_by_context(c.id)

    def get_pipeline_context(self, pipeline_name: Text):
        # We rebuild context for ml metadata here.
        from zenml.core.repo.repo import Repository
        repo: Repository = Repository.get_instance()
        run_id = f'{repo.get_artifact_store().unique_id}.{pipeline_name}'
        run_context = self.store.get_context_by_type_and_name(
            type_name='run',
            context_name=run_id
        )
        if run_context is None:
            raise Exception(f'{pipeline_name} does not exist in Metadata '
                            f'store. This might be due to the fact that it '
                            f'has not run yet!')
        return run_context

    def get_artifacts_by_component(self, pipeline_name: Text,
                                   component_name: Text):
        """
        Args:
            pipeline_name:
            component_name:
        """
        # First , you get the execution associated with the component
        e = self.get_component_execution(pipeline_name, component_name)

        if e is None:
            raise Exception(f'{component_name} not found! This might be due '
                            f'to the fact that the pipeline does not have '
                            f'the associated {component_name} Step.')

        # Second, you will get artifacts
        return self.get_artifacts_by_execution(e.id)

    def get_component_execution(self, pipeline_name: Text,
                                component_name: Text):
        pipeline_executions = self.get_pipeline_executions(pipeline_name)
        for e in pipeline_executions:
            # TODO: [LOW] Create a more refined way to find components.
            if component_name in e.properties['component_id'].string_value:
                return e

    def get_artifacts_by_execution(self, execution_id):
        # First get all the events for this execution
        """
        Args:
            execution_id:
        """
        events = self.store.get_events_by_execution_ids([execution_id])

        artifact_ids = []
        for e in events:
            # MAJOR Assumptions: 4 means its output channel
            if e.type == 4:
                artifact_ids.append(e.artifact_id)
                break

        if not artifact_ids:
            raise AssertionError("This execution has no output artifact.")

        # Get artifacts
        return self.store.get_artifacts_by_id(artifact_ids)

    def get_pipeline_contexts(self) -> List:
        """Gets list of pipeline contexts"""
        return self.store.get_contexts_by_type(type_name='pipeline')

    # old
    def get_run_context_by_comp_execution_id(self, comp_exec_id):
        """
        Args:
            comp_exec_id:
        """
        all_cs = self.store.get_contexts_by_execution(comp_exec_id)
        for c in all_cs:
            if 'run_id' in c.properties:
                return c

    def get_execution_ids_by_value(self, param, value):
        """
        Args:
            param:
            value:
        """
        if self.store:
            db_uri = 'mysql+pymysql://{user}:{password}@{host}/{db}'.format(
                user=self.username,
                password=self.password,
                host=self.host,
                db=self.database,
            )
            engine = create_engine(db_uri)
            with engine.connect() as con:
                query = """
                SELECT * FROM ExecutionProperty
                WHERE (name='{}' and string_value='{}')
                """.format(param, value)
                rs = con.execute(query)
                return [r.execution_id for r in rs]
        else:
            raise AssertionError("Metadata store does not exist.")

    # old stuff
    def visualize_contexts(self):
        """Creates a DataFrame of the available contexts :return:
        pd.DataFrame
        """
        context_df = pd.DataFrame(columns=['run_id',
                                           'pipeline_name',
                                           'context_name',
                                           'context_id'])
        for c in self.store.get_contexts():
            context_df.loc[c.id] = pd.Series({
                'context_id': c.id,
                'context_name': c.name,
                'run_id': c.properties['run_id'].string_value,
                'pipeline_name': c.properties['pipeline_name'].string_value})
        return context_df

    def get_executions_by_context(self, context: str):
        """
        Args:
            context (str):
        """
        context_id = self.store.get_context_by_type_and_name(
            type_name='run',
            context_name=context,
        ).id
        return self.store.get_executions_by_context(context_id)

    def list_registered_components(self):
        """returns a list of registered components :return: list"""
        executions = self.store.get_executions()
        component_list = list({e.properties['component_id'].string_value
                               for e in executions})
        return component_list

    def filter_executions_by_component(self, component):
        """
        Args:
            component:
        """
        assert component in self.list_registered_components()
        executions = self.store.get_executions()
        return [e for e in executions
                if e.properties['component_id'].string_value == component]

    def filter_executions_by_condition(self, condition):
        """
        Args:
            condition:
        """
        component = condition['component']
        property_name = condition['property_name']
        value = condition['value']

        executions = self.filter_executions_by_component(component)
        return [e for e in executions
                if e.properties[property_name].string_value == value]

    def map_execution_to_context(self, executions):
        """
        Args:
            executions:
        """
        return [self.store.get_contexts_by_execution(e.id)[0]
                for e in executions]

    def list_contexts_by_condition(self, condition):
        """
        Args:
            condition:
        """
        executions = self.filter_executions_by_condition(condition)
        contexts = self.map_execution_to_context(executions)
        return contexts

    # Got shit done

    def get_outcomes_in_context(self,
                                context: str = None,
                                output_components: List = None):
        """
        Args:
            context (str):
            output_components (List):
        """
        if output_components:
            output_ids = output_components
        else:
            output_ids = self.list_registered_components()

        context_id = self.store.get_context_by_type_and_name(
            type_name='run',
            context_name=context,
        ).id

        # Among all the executions in context, select the right component
        execs_in_context = self.store.get_executions_by_context(context_id)
        execs_components = {e.properties['component_id'].string_value: e.id
                            for e in execs_in_context
                            if e.properties['component_id'].string_value
                            in output_ids}

        result = {}
        # Get the corresponding event and extract the output artifact ids of it
        for component, exec_id in execs_components.items():
            events = self.store.get_events_by_execution_ids([exec_id])
            artifact_ids = [e.artifact_id for e in events if e.type == 4]
            result[component] = self.store.get_artifacts_by_id(artifact_ids)
        return result

    def get_outcomes(self,
                     contexts: List[Text] = None,
                     context_names: List[Text] = None,
                     output_components: List[Text] = None,
                     conditions: List[Dict] = None):

        """
        Args:
            contexts:
            context_names:
            output_components:
            conditions:
        """
        if bool(contexts) and bool(context_names):
            raise Exception("Cant define context id/names at the same time")

        if contexts:
            target_context_ids = [
                self.store.get_context_by_type_and_name(type_name='run',
                                                        context_name=c).id for
                c in contexts]
        elif context_names:
            target_contexts = self.store.get_contexts()
            target_context_ids = [c.id for c in target_contexts if
                                  c.name in context_names]
        else:
            target_contexts = self.store.get_contexts()
            target_context_ids = [c.id for c in target_contexts]

        if conditions:
            for cond in conditions:
                context_w_cond = self.list_contexts_by_condition(cond)
                context_w_cond_ids = [c.id for c in context_w_cond]
                target_context_ids = [c for c in target_context_ids
                                      if c in context_w_cond_ids]

        result = {}
        for c_id in target_context_ids:
            result[c_id] = self.get_outcomes_in_context(c_id,
                                                        output_components)

        return {k: self.extract_uris(v) for k, v in result.items()}

    def get_context_by_run_id(self, run_id):
        """
        Args:
            run_id:
        """
        contexts = self.store.get_contexts()
        for c in contexts:
            if c.properties['run_id'].string_value == run_id:
                return c.id

    def get_execution_by_component_type(self, context, component_id):
        """
        Args:
            context:
            component_id:
        """
        context_id = self.store.get_context_by_type_and_name(
            type_name='run',
            context_name=context,
        ).id
        execs = self.store.get_executions_by_context(context_id)
        for e in execs:
            full = e.properties['component_id'].string_value
            instance_name = full.split('.')[1]
            if instance_name == component_id:
                return e

    def is_component_cached(self, context, component_id):
        """
        Args:
            context:
            component_id:
        """
        context_id = self.store.get_context_by_type_and_name(
            type_name='run',
            context_name=context,
        ).id
        execs = self.store.get_executions_by_context(context_id)
        for e in execs:
            full = e.properties['component_id'].string_value
            instance_name = full.split('.')[1]
            if instance_name == component_id:
                if e.properties[
                    'state'].string_value == ExecutionStates.cached.name:
                    return True
                else:
                    return False
        return False

    def find_original_context_and_artifacts_from_exec(self, execution):
        """
        Args:
            execution:
        """
        events = self.store.get_events_by_execution_ids([execution.id])

        # get all the OUTPUT artifacts from these events
        artifacts_ids = []
        for e in events:
            if e.type == 4:
                artifacts_ids.append(e.artifact_id)

        if artifacts_ids:
            artifact_id = artifacts_ids[-1]  # lets take the last one

        # If cached
        if execution.properties[
            'state'].string_value == ExecutionStates.cached.name:
            new_events = self.store.get_events_by_artifact_ids([artifact_id])
            for e in new_events:
                if e.type == 4:
                    return self.find_original_context_and_artifacts_from_exec(
                        self.store.get_executions_by_id([e.execution_id])[0])
        else:
            # we got the relevant run
            # run_id = execution.properties['run_id'].string_value
            context = self.store.get_contexts_by_execution(execution.id)[
                0].name

            # we got the relevant artifacts
            artifacts = self.store.get_artifacts_by_id(artifacts_ids)

            # and we got the relevant execution
            return context, artifacts

    def find_original_context_and_artifacts_from_context(self,
                                                         context,
                                                         component_type):
        """
        Args:
            context:
            component_type:
        """
        execution = self.get_execution_by_component_type(
            context,
            component_type
        )
        return self.find_original_context_and_artifacts_from_exec(execution)

    def get_next_context_id(self):
        if self.store:
            db_uri = 'mysql+pymysql://{user}:{password}@{host}/{db}'.format(
                user=self.username,
                password=self.password,
                host=self.host,
                db=self.database,
            )
            engine = create_engine(db_uri)
            with engine.connect() as con:
                query = """
                SELECT AUTO_INCREMENT FROM information_schema.TABLES
                WHERE TABLE_SCHEMA = '{}' AND TABLE_NAME = 'Context'
                """.format(self.database)
                rs = con.execute(query)
                for row in rs:
                    return row[0]
        else:
            raise AssertionError("Metadata store does not exist.")

    @staticmethod
    def extract_uris(result):
        """
        Args:
            result:
        """
        return {k: [a.uri for a in v] for k, v in result.items()}
